from __future__ import annotations

from dataclasses import dataclass
from typing import Union

import numpy as np
from shapely.geometry import Polygon


@dataclass
class Projection:
    plane: str
    polygon: Polygon
    base_vector: list[int]
    anti_vector: list[int]
    angle: int


@dataclass
class Flow:
    num_cylinders: int
    projected_area: np.float64
    surface_area: np.float64
    angle_sum: np.float64
    volume: np.float64
    sa_to_vol: np.float64
    drip_node_id: int
    drip_node_loc: tuple

    def __eq__(self, flow: Flow):
        return (
            np.round(self.num_cylinders, decimals=1)
            == np.round(flow.num_cylinders, decimals=1)
            and np.round(self.projected_area, decimals=3)
            == np.round(flow.projected_area, decimals=3)
            and np.round(self.surface_area, decimals=3)
            == np.round(flow.surface_area, decimals=3)
            and np.round(self.angle_sum, decimals=3)
            == np.round(flow.angle_sum, decimals=3)
            and np.round(self.volume, decimals=3) == np.round(flow.volume, decimals=3)
            and np.round(self.sa_to_vol, decimals=3)
            == np.round(flow.sa_to_vol, decimals=3)
        )


coord_list = Union[list[tuple[np.float64]], np.ndarray]



from __future__ import annotations

import calendar
import csv
import logging
import os
import shutil
import stat
import time
from importlib.util import find_spec
from typing import Callable

import numpy as np

from canopyhydro.configuration import input_dir, output_dir

log = logging.getLogger("model")

current_GMT = time.gmtime()
time_stamp = str(calendar.timegm(current_GMT))


def _try_import(package_name):
    if find_spec(package_name):
        return True
    else:
        return False


has_numba = _try_import("numba")
if has_numba:
    from numba import njit, prange
    from numba.typed import List

# Data munging utils


def stack(to_stack: list[np.array], col: bool = True):
    """
    A wrapper for njit stack that handles errors and allows for
    less strict typing
    """
    if not has_numba:
        non_njit_stack(to_stack, col)
    else:
        list_of_array = List(to_stack)
        try:
            njit_stack(list_of_array, col)
        except ValueError:
            non_njit_stack(to_stack, col)


def non_njit_stack(to_stack: list[np.array], col: bool = True):
    list_of_array = list(to_stack)
    num_in = len(list_of_array)
    left_shape = list_of_array[0].shape[0]
    shape = (num_in, left_shape)
    stacked_array = np.empty(shape)
    for idx, _ in enumerate(list_of_array):
        stacked_array[idx] = list_of_array[idx]
    return stacked_array if not col else stacked_array.T


if has_numba:

    @njit()
    def njit_stack(list_of_array: np.array[np.array()], col: bool):
        """
        numba doesn't play well with np stacks, so I had to do it myself
        """
        num_in = len(list_of_array)
        left_shape = list_of_array[0].shape[0]
        shape = (num_in, left_shape)
        stacked_array = np.empty(shape)
        for j in prange(len(list_of_array)):
            stacked_array[j] = list_of_array[j]
        return stacked_array if not col else stacked_array.T
# File system utils


def on_rm_error(path):
    os.chmod(path, stat.S_IWRITE)
    os.unlink(path)


def create_dir_and_file(
    filename,
) -> None:
    os.makedirs(filename, exist_ok=True)


def del_dir(filename) -> None:
    shutil.rmtree(filename, onerror=on_rm_error)


def read_file_names(file_path=input_dir):
    """Reads in filenames to list"""
    paths = sorted(file_path.iterdir(), key=os.path.getmtime)
    fileNames = [f.name for f in paths if f.suffix == ".csv"]
    log.info(f"fileNames found in {file_path} on read: {fileNames}")
    return fileNames


def save_file(
    file: str,
    out_file: dict | list[dict],
    overwrite: bool = False,
    fileFormat: str = ".csv",
    method: str = "",
):
    """
    A somewhat overly complex file saving function.
    If file exists and overwrite = false, it will append to the file

    Args:
        file: str
            The name of the file to write to
        out_file: Union[dict, list[dict]]
            The data to write to the file
        overwrite: bool
            Whether or not to overwrite the file if it exists
        fileFormat: str
            The file format to save the file as ()
        method: str
    Note:
        'agg*' variables intended to support future
            updates - adding additional write to an
            append only file for better history tracking
    """
    dir = "/".join([str(output_dir), method, ""])
    ofname = "_".join([file, method])
    ofname_ext = "".join([ofname, fileFormat])
    folderExists = os.path.exists(dir)
    fileExists = os.path.exists(dir + ofname_ext)

    # aggname = "_".join(["agg", method, fileFormat])
    # aggname_ext = "".join([aggname, fileFormat])
    # aggExists = os.path.exists(dir + aggname_ext)#

    if not folderExists:
        os.makedirs(dir)

    to_write = []
    if isinstance(out_file, dict):
        out_file = [out_file]

    headers = list(out_file[0].keys())
    to_write.append(headers)
    log.info(f"{to_write}")
    # adding each dict in out_file to a
    #   single to_write list[list]
    for dic in out_file:
        cur_row = []
        for _, value in dic.items():
            cur_row.append(value)
        cur_row.append(time_stamp)
        to_write.append(cur_row)

    if fileExists and not overwrite:
        # Reading existing data into to_write in
        #  order to append new data
        with open(dir + ofname_ext, "w+") as csv_file:
            reader = csv.reader(csv_file)
            existing_rows = list(reader)
            if existing_rows[0] == headers:
                for row in existing_rows[1:]:
                    if row != []:
                        to_write.append(row)
            else:
                log.warning(
                    f"Existing { ofname_ext} file has different headers, to overwrite pass ovewrite =true"
                )
    if overwrite:
        # log.info(f"{to_write}")
        file = dir + ofname_ext
        log.info(f"attempting to write to {file}")
        with open(file, "w") as csv_file:
            writer = csv.writer(csv_file)
            for row in to_write:
                writer.writerow(row)
    return file


# other utils


def intermitent_log(prog: int, whole: int, msg: str, freq: int = 0.0001):
    if np.random.uniform(0, 1, 1) < freq:
        log.info(msg + str(np.round((prog / whole) * 100, decimals=1)))
        print(msg + str(np.round((prog / whole) * 100, decimals=1)))


def lam_filter(objects, a_lambda: Callable, return_all: bool = False):
    """
    Takes in a lambda that filters on cylinder attrs
    returns a list of cylinders for which that
    lambda func returns true
    i.e. lam_filter(cylinders, lambda: diameter > 0.5)
            -returns all cylinders with diameter > 0.5

    return_all: bool
        - if true, returns all cylinders along with
            a boolean array of the same length w/ the
            results of the lambda function passed
    """
    if a_lambda.__code__.co_name != "<lambda>":
        raise ValueError("a lambda required")
    if return_all:
        filtered = [(obj, eval(a_lambda.__code__, vars(obj).copy())) for obj in objects]
    else:
        filtered = [
            (obj, True) for obj in objects if eval(a_lambda.__code__, vars(obj).copy())
        ]
    objs, res = zip(*filtered)
    return objs, res



from __future__ import annotations

# from test.expected_results import ez_projection_vectors, ez_projection_xy_angle
from test.utils import within_range

import numpy as np
import pytest

from canopyhydro.Cylinder import Cylinder

# Test Config
accepted_err = 0.02

# general test_cyl
cyl_list = np.array(
    [
        None,
        0.000,
        -1.000,
        -0.299115,
        2.537844,
        -0.598273,
        -0.299115,
        2.537844,
        -0.552,
        0.557981,
        0.045261,
        7.269781,
        0.046273,
        7955.52145,
        None,
        0.000,
        -1.000,
        0.0529898,
        5.983393,
        7954.0735,
        0.000,
        67.0,
        1.35908057e02,
        18471,
        0.000,
        0.000,
        None,
        0.000,
        0.000,
        0.000,
        0.01,
        0.0401606,
        0.000,
        0.01,
        0.0401606,
    ]
)

cyl_obj_param = Cylinder(
    cyl_id=0.0,
    x=np.array([-0.299115, -0.299115], dtype=object),
    y=np.array([2.537844, 2.537844], dtype=object),
    z=np.array([-0.598273, -0.552], dtype=object),
    radius=0.557981,
    length=0.046273,
    branch_order=0.0,
    branch_id=0.0,
    volume=0.045261,
    parent_id=-1.0,
    reverse_branch_order=67.0,
    segment_id=0.0,
    projected_data={},
    flow_id=None,
    flow_type=None,
    begins_at_drip_point=None,
    begins_at_divide_point=None,
    dx=0.0,
    dy=0.0,
    dz=0.04627300000000001,
    surface_area=0.16222841912042885,
    sa_to_vol=3.584287115185896,
    slope=0.0,
    is_stem=False,
    is_divide=False,
)

cyl_sa = 0.162

# projection_test_cyl
proj_cyl_list = np.array(
    [
        None,
        0.00000000e00,
        -1.00000000e00,
        1.00000000e00,
        1.00000000e00,
        1.00000000e00,
        4.00000000e00,
        6.00000000e00,
        7.00000000e00,
        1.00000000e00,
        1.00210000e-02,
        3.09237800e00,
        6.44330000e-02,
        2.13541289e03,
        None,
        0.00000000e00,
        -1.00000000e00,
        1.97783000e-01,
        2.94657100e00,
        2.13431216e03,
        0.00000000e00,
        3.20000000e01,
        5.67450440e01,
        3.22000000e03,
        0.00000000e00,
        0.00000000e00,
        None,
        0.00000000e00,
        0.00000000e00,
        0.00000000e00,
        1.00000000e-02,
        4.01606000e-01,
        0.00000000e00,
        1.00000000e-02,
        4.01606000e-01,
    ]
)

proj_cyl = Cylinder(
    cyl_id=0.0,
    x=np.array([1.0, 4.0]),
    y=np.array([1.0, 6.0]),
    z=np.array([1.0, 7.0]),
    radius=1.0,
    length=0.064433,
    branch_order=0.0,
    branch_id=0.0,
    volume=0.010021,
    parent_id=-1.0,
    reverse_branch_order=32.0,
    segment_id=0.0,
    projected_data={},
    flow_id=None,
    flow_type=None,
    begins_at_drip_point=None,
    begins_at_divide_point=None,
    dx=3.0,
    dy=5.0,
    dz=6.0,
    surface_area=0.40484447889750186,
    sa_to_vol=40.399608711456125,
    slope=0.0,
    is_stem=False,
    is_divide=False,
)

ez_projection_vectors = {
    "XY": [np.array([1.0, 1.0, 1.0]), np.array([4.0, 6.0, 7.0])],
    "XZ": [np.array([1.0, 1.0, 1.0]), np.array([4.0, 7.0, 6.0])],
    "YZ": [np.array([1.0, 1.0, 1.0]), np.array([6.0, 7.0, 4.0])],
}
ez_projection_xy_angle = 0.7996
ez_projection_xz_angle = 0.6405
ez_projection_yz_angle = 0.3667


@pytest.mark.parametrize(
    "test_cyl, cyl_obj", [(cyl_list, cyl_obj_param)], indirect=["test_cyl"]
)
def test_create_test_cyl(test_cyl, cyl_obj):
    """
    Tests the creation of the test cylinder fixture
    """
    assert test_cyl == cyl_obj


@pytest.mark.parametrize(
    "test_cyl, ex_surface_area", [(cyl_list, cyl_sa)], indirect=["test_cyl"]
)
def test_calc_surface_area(test_cyl, ex_surface_area):
    surface_area = test_cyl.calc_surface_area()
    assert within_range(ex_surface_area, surface_area, accepted_err)


@pytest.mark.parametrize("proj_cyl", [proj_cyl])
def test_project_cylinder(proj_cyl):
    proj_cyl.get_projection("XY")
    actual = proj_cyl.projected_data["XY"]["angle"]
    expected = ez_projection_xy_angle
    assert within_range(expected, actual, accepted_err)


# @pytest.mark.parametrize("proj_cyl", [proj_cyl])
# def test_project_cylinder_numba(proj_cyl):
#     proj_cyl.numba_get_projection("XY")
#     actual = proj_cyl.projected_data["XY"]["angle"]
#     expected = ez_projection_xy_angle
#     assert within_range(expected, actual, accepted_err)


@pytest.mark.parametrize("test_cyl", [(proj_cyl_list)], indirect=["test_cyl"])
def test_create_cyl_angle(test_cyl):
    """
    Tests the creation of the test cylinder fixture
    """
    actual = test_cyl.angle
    expected = ez_projection_xy_angle
    assert within_range(expected, actual, accepted_err)


@pytest.mark.parametrize("test_cyl", [(proj_cyl_list)], indirect=["test_cyl"])
def test_create_cyl_vectors(test_cyl):
    """
    Tests the creation of the test cylinder fixture
    """
    actual = {
        "XY": [
            np.np.array([test_cyl.x[0], test_cyl.y[0], test_cyl.z[0]]),
            np.np.array([test_cyl.x[1], test_cyl.y[1], test_cyl.z[1]]),
        ],
        "XZ": [
            np.np.array([test_cyl.x[0], test_cyl.z[0], test_cyl.y[0]]),
            np.np.array([test_cyl.x[1], test_cyl.z[1], test_cyl.y[1]]),
        ],
        "YZ": [
            np.np.array([test_cyl.y[0], test_cyl.z[0], test_cyl.x[0]]),
            np.np.array([test_cyl.y[1], test_cyl.z[1], test_cyl.x[1]]),
        ],
    }
    expected = ez_projection_vectors
    # Probably should mae a  data class for these cylinder vectors... oh well
    for k, v in actual.items():
        for idx, vector in enumerate(v):
            assert np.all(expected[k][idx] == vector)



from __future__ import annotations

import os
import shutil
from pathlib import Path

import pytest
from numpy import all
from numpy import array as arr

from canopyhydro.configuration import root_dir as DIR
from canopyhydro.utils import (create_dir_and_file, njit_stack, on_rm_error,
                               read_file_names, stack)

njit_stack_test_cases = [
    pytest.param(
        arr([arr([1, 2, 3]), arr([4, 5, 6])]),
        False,
        arr([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]),
        id="3x3 row stack",
    ),
    pytest.param(
        arr([arr([1, 2, 3]), arr([4, 5, 6]), arr([7, 8, 9])]),
        False,
        arr([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]),
        id="3x3x3 row stack",
    ),
    pytest.param(
        arr([arr([1, 2]), arr([4, 5])]),
        False,
        arr([[1.0, 2.0], [4.0, 5.0]]),
        id="2x2 row stack",
    ),
    pytest.param(
        arr([arr([1, 2, 3]), arr([4, 5, 6])]),
        True,
        arr([[1.0, 4.0], [2.0, 5.0], [3.0, 6.0]]),
        id="3x3 col stack",
    ),
    pytest.param(
        arr([arr([1, 2]), arr([4, 5])]),
        True,
        arr([[1.0, 4.0], [2.0, 5.0]]),
        id="2x2 col stack",
    ),
]

stack_test_cases = [
    pytest.param(
        [arr([1]), arr([4, 5]), arr([4, 5])], False, ValueError, id="1x2x2 ValueError"
    ),
    pytest.param([arr([1]), arr([4, 5])], False, ValueError, id="1x2 ValueError"),
]


def test_file_funcs():
    file_path = "".join([DIR, "test_data/"])
    file = os.path.dirname(file_path)
    create_dir_and_file(file)
    f = open("test_data/demofile2.csv", "w")
    f.write("Now the file has more content!")
    f.close()
    file_names = read_file_names(Path("".join([DIR, "test_data"])))
    assert file_names == ["demofile2.csv"]
    shutil.rmtree(file, onerror=on_rm_error)
    print("File Names Successful")


@pytest.mark.parametrize("arr_list,is_col_stack,expected", njit_stack_test_cases)
def test_njit_stack(arr_list, is_col_stack, expected):
    """
    Test vstack and col stack capabilities v. numpy
    """
    actual = njit_stack(arr_list, is_col_stack)
    assert all(actual == expected)


@pytest.mark.parametrize("arr_list,is_col_stack,exception_type", stack_test_cases)
def test_stack_err(arr_list, is_col_stack, exception_type):
    """
    Test vstack and col stack capabilities v. numpy
    """
    with pytest.raises(exception_type):
        stack(arr_list, is_col_stack)



from __future__ import annotations

import logging
import logging.config
import os

import toml
import yaml

# Read in environment variables, set defaults if not present

config_file = os.environ.get("CANOPYHYDRO_CONFIG", "./canopyhydro_config.toml")
log_config = os.environ.get("CANOPYHYDRO_LOG_CONFIG", "./logging_config.yml")

with open(log_config) as f:
    config = yaml.safe_load(f.read())
    logging.config.dictConfig(config)

log = logging.getLogger(__name__)

in_flow_grade_lim = None
output_dir = None
input_dir = None
qsm_cols = {}

try:
    with open(config_file) as f:
        config = toml.load(f)
        in_flow_grade_lim = config["model_parameters"]["in_flow_grade_lim"]
        root_dir = config["directories"]["root_dir"]
        output_dir = config["directories"]["output_dir"]
        input_dir = config["directories"]["input_dir"]
        test_input_dir = config["directories"]["test_input_dir"]
        for column in config["qsm"]:
            qsm_cols[column] = config["qsm"][column]
except Exception as e:
    log.error(f"Error loading configuration variables from {config_file}: {e}")
    raise e



"""Spacially centeric code"""
from __future__ import annotations

import functools
import logging
import math

import numpy as np
from scipy.linalg import lu_factor, lu_solve
from scipy.spatial import Delaunay, distance
from shapely.geometry import MultiLineString, MultiPoint, MultiPolygon, Polygon
from shapely.ops import polygonize, unary_union

from canopyhydro.configuration import output_dir
from canopyhydro.DataClasses import coord_list
from canopyhydro.utils import _try_import

log = logging.getLogger("model")

if has_geopandas := _try_import("geopandas"):
    from geopandas import GeoSeries


if has_matplotlib := _try_import("matplotlib"):
    import matplotlib.pyplot as plt
    from matplotlib.collections import PatchCollection
    from matplotlib.colors import to_rgba
    from matplotlib.patches import PathPatch
    from matplotlib.path import Path

    # Plots a Polygon to pyplot `ax`
    def polygon_plot(ax, polys: list | Polygon, **kwargs):
        if isinstance(polys, Polygon):
            polys = [polys]
        if isinstance(polys, MultiPolygon):
            polys = list(polys.geoms)
        colors = kwargs.get("facecolor", ["grey"])
        if isinstance(colors, str):
            colors = list(colors)
        union_dict = {color: [] for color in colors}
        for poly, color in zip(polys, colors):
            union_dict[color].append(poly)
        for color, polys in union_dict.items():
            if polys != []:
                to_draw = unary_union(polys)
                kwargs["facecolor"] = to_rgba(color)

                path = Path.make_compound_path(
                    Path(np.asarray(to_draw.exterior.coords)[:, :2]),
                    *[
                        Path(np.asarray(ring.coords)[:, :2])
                        for ring in to_draw.interiors
                    ],
                )

                patch = PathPatch(path, **kwargs)
                collection = PatchCollection([patch], **kwargs)
                ax.add_collection(collection, autolim=True)
                ax.autoscale_view()
        return collection
else:

    def polygon_plot(ax, poly, **kwargs):
        raise ImportError("matplotlib is required to plot polygons")


### Functions relating to Alpha Shapes
def circumcenter_lapack(points: coord_list) -> np.ndarray:
    """
    Calculate the circumcenter of a set of points relative to simplex
    https://en.wikipedia.org/wiki/Polarization_identity
    """
    points = np.asarray(points)
    rows, _ = points.shape
    A = np.bmat(
        obj=[
            [2 * np.dot(points, points.T), np.ones((rows, 1))],
            [np.ones((1, rows)), np.zeros((1, 1))],
        ]
    )
    b = np.hstack((np.sum(points * points, axis=1), np.ones(1)))
    return np.array(np.linalg.solve(A, b)[:-1])


def circumcenter_lu_factor(points: coord_list) -> np.ndarray:
    """
    Calculate the circumcenter of a set of points relative to simplex
    Theoretically less efficient than LAPACK (O(n^3) + O(n^2) v. O(n^2))
        when a given set of equations (e.g. set of points, A )
        only needs to be solved for a single vector (e.g b)
    """
    points = np.asarray(points)
    rows, _ = points.shape
    A = np.bmat(
        obj=[
            [2 * np.dot(points, points.T), np.ones((rows, 1))],
            [np.ones((1, rows)), np.zeros((1, 1))],
        ]
    )
    b = np.hstack((np.sum(points * points, axis=1), np.ones(1)))
    lu, piv = lu_factor(A)
    sir_c = lu_solve((lu, piv), b)
    return sir_c


def circumradius(points: coord_list, center: np.ndarray == []) -> np.float32:
    """
    Calculte the radius of the circle in which the given polygon may be inscribed
    """
    points = np.asarray(points)
    return np.linalg.norm(points[0, :] - np.dot(center, points))


def simplices(points: coord_list) -> coord_list:
    """
    Yields simpicies and radius
    """
    coords = np.asarray(points)
    tri = Delaunay(coords)

    for simplex in tri.simplices:
        simplex_points = coords[simplex]
        try:
            center = circumcenter_lapack(simplex_points)
            yield simplex, circumradius(simplex_points, center), center
        except np.linalg.LinAlgError as err:
            log.warning(
                f"""Error calculating simplicies,
                        input invalid (potentially coincident points)
                         {err}"""
            )


def maximal_alpha(boundary_points: coord_list, union_poly: Polygon) -> np.float32:
    """
    Finds the minimal alpha shape for the given
    coord list that still contains the given polygon
    """
    upper = (
        10  # annectotally seems to be plenty high to ensure a discontinuous alpha shape
    )
    lower = 0
    iterations = 10
    runs = 0
    while runs <= iterations:
        alpha = (upper - lower) / 2
        hull, _ = concave_hull(boundary_points, union_poly)
        if hull.contains(Polygon):
            # The tested curvature is less than or equal to the max curvature for a continuous alpha shape
            lower = alpha
        else:
            # The tested curvature is greater than or equal to the max curvature for a continuous alpha shape
            upper = alpha
        runs += 1
    return alpha


# @profile
# might eventually be updated to deal
# with 3d a la https://github.com/bellockk/alphashape/blob/master/alphashape/optimizealpha.py
def concave_hull(boundary_points, alpha: int = 0):
    """alpha shape / concave hull
    Creates a delaunay triangulation on the set of points. Using this triangulation,
    returns a minimal concave polygon with a concavity factor alpha
    see: trunk_lean
    """

    if len(boundary_points) < 4:
        # When you have a triangle, there is no sense in computing an alpha
        # shape.
        return MultiPoint(list(boundary_points)).convex_hull, boundary_points

    def add_edge(edges, edge_points, coords, i, j):
        # adds a line between points i and j
        if (i, j) in edges or (j, i) in edges:
            # already added
            return
        edges.add((i, j))
        edge_points.append(coords[[i, j]])

    def add_center(centers, center):
        centers.add((center[0], center[1], center[2]))

    coords = np.array(
        sorted(point.coords[0] for point in boundary_points if len(point.coords) > 0)
    )

    # Find minimal set of triangles with points in set

    edges = set()
    centers = set()
    edge_points = []
    # loop over triangles:
    # ia, ib, ic = indices of corner points of the triangle
    tri = simplices(coords)

    for (ia, ib, ic), _, center in tri:
        if len(coords[0]) == 2:
            # simpler logic suffices in 2d
            pa = coords[ia]
            pb = coords[ib]
            pc = coords[ic]

            # Lengths of sides of triangle
            a = math.sqrt((pa[0] - pb[0]) ** 2 + (pa[1] - pb[1]) ** 2)
            b = math.sqrt((pb[0] - pc[0]) ** 2 + (pb[1] - pc[1]) ** 2)
            c = math.sqrt((pc[0] - pa[0]) ** 2 + (pc[1] - pa[1]) ** 2)

            # Semiperimeter of triangle
            s = (a + b + c) / 2.0

            # Area of triangle by Heron's formula
            area = math.sqrt(s * (s - a) * (s - b) * (s - c))
            circum_r = a * b * c / (4.0 * area)

            # Here's the radius filter.
            # print circum_r
            if circum_r < 1.0 / alpha:
                add_edge(edges, edge_points, coords, ia, ib)
                add_edge(edges, edge_points, coords, ib, ic)
                add_edge(edges, edge_points, coords, ic, ia)
                add_center(centers, center)
        else:
            raise (Exception("3D alpha shapes not yet supported"))
    m = MultiLineString(edge_points)
    triangles = list(polygonize(m))
    return unary_union(triangles), edge_points, centers


# def draw_vectors(vectors: list[tuple[list[int]]],
#                  colors: Optional[list[str]] = ["black"],
#                  save: bool = False,
#                  show: bool = False,
#                  file_name_ext: str = "",
#                  )->tuple[int,int,int]:
#     """Gets the x, y and z coordinates from a list of vectors
#         as separate lists.

#     ### Args:
#         - `vectors (list[tuple[list[int]]])`: _description_

#     ### Returns:
#         - `_type_`: _description_
#     """
#     for x,y,z in vectors:
#         plt.plot(x, y, z, color="black")
#     if save:
#         save_dir = "/".join(
#             [str(output_dir), "draw_vectors", f"{file_name_ext.replace('.','')}"]
#         )  # .replace("/", "\\")
#         plt.savefig(save_dir, dpi=3000)
#     if show:
#         plt.show()


### Functions relating to 2D Cylinder Projections
def get_projection(vector: list, magnitude: list, radius: float()):
    """
    Takes in the vector (starting point), magnitude and radius that fully define a cylinder.
    Finds the projection of the cylinder on a plane

    Some linear algebra/diff eq could help us find this for an arbitrary plane.
    """
    noCirPoints = 360
    tCir = np.linspace(
        0, 2 * np.pi, noCirPoints
    )  # 360 evenly spaced points between 0 - 2pi (radian degrees)
    a_ortho = np.cos(tCir)  # x coordinates of the points on a circle
    b_ortho = np.sin(tCir)  # y coordinates of the points on a circle
    delt_a = magnitude[0]
    delt_b = magnitude[1]
    delt_c = magnitude[2]
    dim_a = vector[0]
    dim_b = vector[1]
    # unit vector at base of cylinder, pointing up cylinder axis
    vNorm = np.sqrt(delt_a**2 + delt_b**2 + delt_c**2)
    aV = np.hstack((delt_a, delt_b, delt_c)) / vNorm
    bV = -aV  # unit vector looking down from top circle (but not translated)
    # function to find the z component of an orthogonal vector in 3D
    # oVz = lambda v, a, b: ((-v[0] * a - v[1] * b) / v[2])

    # initializing min max arrays+
    projection = {
        "polygon": Polygon(),
        "base_vector": (0, 0, 0),
        "anti_vector": (0, 0, 0),
        "angle": 0,
        "area": 0,
    }
    c1 = Polygon()
    bBox = Polygon()
    c2 = Polygon()
    # coord_list = []
    try:
        # for each cylinder
        if not np.isnan(dim_a[0]):
            if np.logical_and(delt_a == 0, delt_b == 0):
                pX = dim_a[0] + radius * a_ortho
                pY = dim_b[0] + radius * b_ortho
                cPS = Polygon(list(zip(pX, pY)))
                ang = 0
                projection = {
                    "polygon": cPS,
                    "base_vector": aV,
                    "anti_vector": bV,
                    "angle": ang,
                    "area": cPS.area,
                }
                return projection
            else:
                # find orthogonal unit vectors @ endpoints
                # Identifies corners of projected rectangle
                # Recall that aV is a unit vector pointing 'up' the cylinder
                aVp1 = np.hstack((aV[1], -aV[0]))
                aVp2 = np.hstack((-aV[1], aV[0]))
                bVp1 = np.hstack((bV[1], -bV[0]))
                bVp2 = np.hstack((-bV[1], bV[0]))
                aVp1 = aVp1 / np.linalg.norm(aVp1)
                aVp2 = aVp2 / np.linalg.norm(aVp2)
                bVp1 = bVp1 / np.linalg.norm(bVp1)
                bVp2 = bVp2 / np.linalg.norm(bVp2)
                # from each endpoint, use radius to find points on the edge of
                # the circle representing the end of the cylinder
                # vertices of the rectangle
                x1 = dim_a[0] + radius * aVp1[0]
                y1 = dim_b[0] + radius * aVp1[1]
                x2 = dim_a[0] + radius * aVp2[0]
                y2 = dim_b[0] + radius * aVp2[1]
                x3 = dim_a[1] + radius * bVp1[0]
                y3 = dim_b[1] + radius * bVp1[1]
                x4 = dim_a[1] + radius * bVp2[0]
                y4 = dim_b[1] + radius * bVp2[1]

                if aV[2] != 0:
                    # calculate set of orthgonal vectors using lambda function
                    # That is 360 orthogonal vectors ending at eqidistant points along
                    # a circle of radius radius with the starting point of our cylinder
                    # at is center
                    ZOrtho = (-aV[0] * a_ortho - aV[1] * b_ortho) / aV[2]
                    # unit-ify the orthgonal vectors
                    uovd = np.sqrt(a_ortho**2 + b_ortho**2 + ZOrtho**2)
                    uov = (
                        np.hstack((a_ortho[:, None], b_ortho[:, None], ZOrtho[:, None]))
                        / uovd[:, None]
                    )
                    # donot re unit-fy, you only want the horizontal component, not the
                    # renormalized horizontal component
                    # using only the X and Y components, find circle coods in plane of
                    # interest
                    xaC = dim_a[0] + uov[:, 0] * radius
                    yaC = dim_b[0] + uov[:, 1] * radius
                    # zaC = dim_c[0] + uov[:, 2] * radius
                    xbC = dim_a[1] + uov[:, 0] * radius
                    ybC = dim_b[1] + uov[:, 1] * radius
                    # zbC = dim_c[1] + uov[:, 2] * radius
                    try:
                        c1c = list(
                            zip(
                                [0 if np.isnan(x) else x for x in xaC],
                                [0 if np.isnan(y) else y for y in yaC],
                            )
                        )
                        c2c = list(
                            zip(
                                [0 if np.isnan(x) else x for x in xbC],
                                [0 if np.isnan(y) else y for y in ybC],
                            )
                        )

                        # coord_list.extend(c1c)
                        # coord_list.extend(c2c)
                        c1 = Polygon(c1c)
                        c2 = Polygon(c2c)
                    except Exception as e:
                        log.error(
                            f"Error creating circular portions of the projections {e}"
                        )

                # assemble total package
                rX = np.vstack((x1, x2, x3, x4))
                rY = np.vstack((y1, y2, y3, y4))
                # test for circle parts in polygon
                try:
                    bBoxc = list(
                        zip(
                            [0 if np.isnan(x) else x for x in rX],
                            [0 if np.isnan(y) else y for y in rY],
                        )
                    )
                    # coord_list.extend(bBoxc)
                    bBox = Polygon(bBoxc)
                    partsPS = [c1, bBox, c2]
                except Exception as err:
                    log.error(
                        f"Error creating rectangular portion of the projection: {err=}: {vector=},{magnitude=}"
                    )
                try:
                    if np.max([poly_part.area for poly_part in partsPS]) > 0:
                        to_union = [poly for poly in partsPS if poly.area > 0]
                        cPS = unary_union([part for part in to_union])
                        # cPSc = Polygon(coord_list)
                except Exception as err:
                    log.error(f"Error unioning projection polygons: {err=} ")
                # get angle away from plane projected on to
                run = math.sqrt(delt_b**2 + delt_a**2)
                rise = delt_c
                if run == 0:
                    slope = 1  # straightDown e.g. is in flow
                else:
                    slope = rise / run
                ang = np.arctan(slope)
                projection = {
                    "polygon": cPS,
                    "base_vector": aV,
                    "anti_vector": bV,
                    "angle": ang,
                    "area": cPS.area,
                }
                return projection
        else:
            log.warning("dim_a[0] is null, unable to project")
        return projection
    except UnboundLocalError:
        log.error(
            f"UnboundLocalError: vector : {vector} magnitude: {magnitude} radius: {radius}"
        )


def get_projected_overlap(shading_poly_list: list[list[Polygon]], labels: list) -> dict:
    """Takes in a list of lists of polygons, w/
     each list representing a diff percentile
     grouping of polygons (e.g. grp1:(0%-25%), grp2:(25%, 50%)...)

    Consider the case in which our percentiles are defined by
        height - with the first grouping being cylinders in the xth %ile
        by height, thus being under than cylinders in all of the other groupings
    In this case, this function calculates the area of shade cast by each
        grouping on the below sections. As such, in each loop 'shading_poly_list'
        is partitioned into 2 categories: The sections of canopy on which shade is
        being cast and the sections of canopy that are casting that shade.
        After each loop, a new section of canopy is moved from the 'shading' group
        to the 'shaded' group and a new calculation of shaded area is made. The result
        is a cumulative sum of shaded area at various heights in the canopy.

    Note:
        shapely's intersection function could be used, and
            would be slightly more accurate. However, it is also
            rather slow for the intersection of this many shapes

     Returns:
         list[dict]: Each dicitonary corresponding to a percentile,
         with metrics describing the overlap between the polygons in that percentile
         {
             "sum_area": sum of areas of the polygons in the percentile,
             "effective_area": the area of the union of the polygons in the percentile,
             "internal_overlap": sum_area - effective_area,
             "overlap_with_previous": the area of the overlap between the polygons in the percentile,
                     with the polygons in the previous percentile,
        }
    """
    if len(labels) != len(shading_poly_list):
        log.info(
            f"Not enough labels; expected {len(shading_poly_list)} got {len(labels)}"
        )
    elif len(set(labels)) != len(labels):
        log.info("Labels must be distinct")
    else:
        overlap_dict = {}
        shaded_polys = []
        for idx, shader_polys in enumerate(shading_poly_list):
            shader_union_poly = unary_union(shader_polys)
            shader_sum = np.sum([poly.area for poly in shader_polys])
            shaded_union = unary_union(shaded_polys)
            total_union = unary_union(shaded_polys)

            shader_on_shaded_w_overlap = shader_union_poly.area + shaded_union.area
            shader_on_shaded_w_o_overlap = total_union.area
            shader_on_shaded_overlap = (
                shader_on_shaded_w_overlap - shader_on_shaded_w_o_overlap
            )

            shader_internal_overlap = shader_sum - shader_union_poly.area

            overlap_dict[labels[idx]] = {
                "sum_area": shader_sum,
                "effective_area": shader_union_poly.area,
                "internal_overlap": shader_internal_overlap,
                "overlap_with_previous": shader_on_shaded_overlap,
            }
            shaded_polys.extend(shader_polys)
        return overlap_dict


### Functions relating to 3D Cylinder Calculations
def get_rotation_matrix(b: np.array):
    if np.linalg.norm(b) == 0:
        return np.eye(3)
    # Sourced from https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula#Matrix_notation
    # b must be unit vector
    # a is the z unit vector
    a = [0, 0, 1]
    v = np.cross(b, a)
    s = np.linalg.norm(v)
    c = np.dot(b, a)
    # The skew-symmetric cross product matrix of v
    vx = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    # Rotation matrix as per Rodregues formula
    R = np.eye(3) + vx + np.dot(vx, vx) * ((1 - c) / (s**2))
    return R


@functools.lru_cache
def get_unoriented_cylinder(r, h, noCirPoints=20, nv=20):
    """
    Given the radius (r) and height (h), returns the parameterization of a cylinder
        centered on the z axis, with a base at the origin and oriented vertically
    """
    theta = np.linspace(0, 2 * np.pi, noCirPoints)
    v = np.linspace(0, h, nv)
    theta, v = np.meshgrid(theta, v)
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    z = v
    return x, y, z


def get_cylinder_surface(radius, vector_start_end):
    """Returns points on the surface of the cylinder defined
    by the givent start and end points having the given radius

    """
    vector = vector_start_end[1] - vector_start_end[0]
    vector = np.array(vector)
    length = np.linalg.norm(vector)
    vu = vector / np.linalg.norm(vector)

    # Creates cylinder centered on the z axis with the base at the origin
    Xc, Yc, Zc = get_unoriented_cylinder(radius, length)

    # Creates matrix to rotate cylinder to the desired orientation
    R = get_rotation_matrix(vu)
    t = np.transpose(np.array([Xc, Yc, Zc]))
    x, y, z = np.transpose(t @ R, (2, 0, 1))

    # shifts now oriented cylinder to the desired position
    x = x + vector_start_end[0][0]
    y = y + vector_start_end[0][1]
    z = z + vector_start_end[0][2]
    return x, y, z, vu


## Functions relating to Drawing Cylinders
def draw_cyls(
    collection: list,
    colors: list[bool] = [True],
    save: bool = False,
    file_name_ext: str = "",
    overlay: list | None = None,
    overlay_transparency: float | None = None,
    show: bool = False,
):
    """Draws a collection of cylinders in 2 dimensions"""
    log.info("Plotting cylinder collection")

    fig, ax = plt.subplots()
    geoPolys = GeoSeries(collection)

    colors = ["Blue" if col else "Grey" for col in colors]
    geoPolys.plot(ax=ax, color=colors)
    # polygon_plot(ax, collection, facecolor=colors)
    if overlay:
        for item in overlay:
            # The below if statements allow for the overlay to be:
            #   a single polygon or a list of polygons
            #       - in the case of watershed alpha shapes
            #   or a list of coordinate lists
            #       - in the case of drip points
            if isinstance(item, Polygon) or isinstance(item, MultiPolygon):
                polygon_plot(ax, item, facecolor=["grey"], alpha=overlay_transparency)
            if isinstance(item, list):
                if isinstance(item[0], Polygon):
                    polygon_plot(
                        ax,
                        item,
                        facecolor=["grey"] * len(item),
                        alpha=overlay_transparency,
                    )
                elif isinstance(item[0], list):
                    ax.scatter(*item, alpha=overlay_transparency, color="red")
            else:
                log.error(
                    "Overlay must be a Polygon, a list of Polygons or coordinate list"
                )
    if show:
        plt.show()
    if save:
        save_dir = "/".join(
            [str(output_dir), "draw", f"{file_name_ext.replace('.','')}"]
        )  # .replace("/", "\\")
        plt.savefig(save_dir, dpi=3000)
    return ax


def draw_cylinders_3D(
    radii: int,
    vector_start_ends: list,
    save: bool = False,
    show: bool = False,
    draw_vectors=False,
    draw_projections=False,
    file_name_ext: str = "3Ddraw",
):
    """Draws a single Cylinder in 3 dimensions"""
    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    for radius, vector_start_end in zip(radii, vector_start_ends):
        log.info("Plotting cylinder in 3D")

        x, y, z, unit_vector = get_cylinder_surface(radius, vector_start_end)

        # setting up chart
        xlim = (np.min(x), np.max(x))
        ylim = (np.min(y), np.max(y))
        zlim = (np.min(z), np.max(z))

        ax.plot_surface(x, y, z, alpha=0.6)
        vector = vector_start_end[1] - vector_start_end[0]

        if draw_vectors:
            ax.quiver(
                vector_start_end[0][0],
                vector_start_end[0][1],
                vector_start_end[0][2],
                vector[0],
                vector[1],
                vector[2],
                color="black",
            )

        if draw_projections:
            axis_extension = 4
            xlim = (np.min(x) - axis_extension, np.max(x) + axis_extension)
            ylim = (np.min(y) - axis_extension, np.max(y) + axis_extension)
            zlim = (np.min(z) - axis_extension, np.max(z) + axis_extension)

            # In the figure created, the viewer is on the positive
            # side of the x and z axis, but the  negative side of the y axis
            # So, we need to offset the y (e.g. XZ projection) based on its 2nd coordinate
            ax.contourf(x, y, z, zdir="x", offset=xlim[0], colors="C0")
            ax.contourf(x, y, z, zdir="y", offset=ylim[1], colors="C0")
            ax.contourf(x, y, z, zdir="z", offset=zlim[0], colors="C0")

        # setting viewer perspective on chart
    ax.view_init(elev=30, azim=-45, roll=0)
    ax.set(xlim=xlim, ylim=ylim, zlim=zlim, xlabel="X", ylabel="Y", zlabel="Z")
    if show:
        plt.show()
    if save:
        save_dir = "/".join(
            [str(output_dir), "draw", f"{file_name_ext.replace('.','')}"]
        )  # .replace("/", "\\")
        plt.savefig(save_dir, dpi=3000)
    return fig


### Functions relating to Points in Space
def closest_points(point: tuple, points: np.array, num_returned: int = 3):
    """
    Finds the closest point in the list 'points' from the input 'point'
    """
    points_to_return = []
    distances = distance.cdist([point], points)
    for i in np.arange(num_returned):
        closest_index = distances.argmin()
        points_to_return.append(points[closest_index])
        distances = np.delete(distances, closest_index)

    return points_to_return[0] if num_returned == 1 else points_to_return


def furthest_point(
    point: tuple,
    points: np.array,
):
    """
    Finds the furthest point in the list 'points' from the input 'point'
    """
    distances = distance.cdist([point], points)
    furthest_index = distances.argmax()
    return points[furthest_index], distances[furthest_index]


# def drip_plot(**args):
#     plt.imshow(**args)



from __future__ import annotations

import copy
import logging
import os
import pickle
from io import TextIOWrapper
from itertools import chain
from typing import Callable

import networkx as nx
# import rustworkx as rx
import numpy as np
from shapely.geometry import Point
from shapely.ops import unary_union

from canopyhydro.configuration import in_flow_grade_lim, input_dir, output_dir
from canopyhydro.Cylinder import Cylinder, create_cyl
from canopyhydro.DataClasses import Flow
from canopyhydro.geometry import (concave_hull, draw_cylinders_3D, draw_cyls,
                                  furthest_point, get_projected_overlap)
from canopyhydro.utils import (create_dir_and_file, intermitent_log,
                               lam_filter, save_file)

log = logging.getLogger("model")

NAME = "CylinderCollection"


def pickle_collection(collection: CylinderCollection, file_name_ext: str = ""):
    """
    Pickles a collection to a file so that it may be
        read into a CylinderCollection object at a later time.

    Args:
        collection:
            The collection to pickle.
        file_name_ext:
            the naming convention to used for the saved file name
            i.e. For a collection w/ file name 'myQSM.csv' and file_name_ext
                'test_run' the file created will be 'myQSM_test_run.pickle'
    Example:
        For a collection w/ file name 'myQSM.csv' and file_name_ext
            'test_run' the file created will be 'myQSM_test_run.pickle'

    Returns:
        file_path (str): The path and file name for the the pickled file.

    Raises:
        FileNotFoundError: If the file does not exist.
        pickle.UnpicklingError: If there is an error while unpickling the file.
    """
    qsm_file = f'{collection.file_name.replace(".csv","")}'
    if ".pickle" not in file_name_ext:
        file_name_ext = f"{file_name_ext}.pickle"
    file_to_create = f"{qsm_file}_{file_name_ext}"
    file_path = "".join([output_dir, "pickle/", file_to_create])
    directory = os.path.dirname(file_path)
    create_dir_and_file(directory)
    pickle_file = open(file_path, "ab")
    pickle.dump(collection, pickle_file)
    pickle_file.close()
    return file_path


def unpickle_collection(file_name: str) -> CylinderCollection:
    """
    Unpickles a collection from a file.
    File assumed to be generated by 'pickle_collection' method

    Args:
        file_name (str):
            The name of the file to unpickle. Optionally, the full path to the file.

    Returns:
        db: The unpickled collection.

    Raises:
        FileNotFoundError: If the file does not exist.
        pickle.UnpicklingError: If there is an error while unpickling the file.
    """
    # In the event that the passed file name is
    if "pickle/" in file_name:
        file_path = file_name
    else:
        file_path = "".join([output_dir, "pickle/", file_name])
    dbfile = open(file_path, "rb")
    db = pickle.load(dbfile)
    dbfile.close()
    return db


class CylinderCollection:
    """
    CylinderCollections represent QSM's or sections thereof, but with additional structure.
    A CylinderCollection is initialized from a .csv file and can:
        - Project cylinders onto the XY, XZ, and YZ planes
        - Determine which precipitation partition water intercepted by
            each of its cylinders will flow into.
        - Find common tree characteristics: DBH, trunk lean, etc.
        - Calculate occlusion for given cross sections of the canopy
        - Draw detailed figures representing the canopy watershed, with:
            - Cylinders in 3D or projected on to the XY, XZ, or YZ planes
            - Drip points (areas of dense throughfall)
            - Alpha shapes (concave hulls) around arbitrary parts of the canopy

    Attributes:
       file (str): The path to the .csv file used to initialize the CylinderCollection. Defaults to "".
       surface_area (float):
          The total surface area of all cylinders in the collection. Defaults to np.nan.
       file_name (string):
          The name of the .csv file used to initialize the CylinderCollection. Defaults to "".
       volume (float):
          The total volume of all cylinders in the collection. Defaults to np.nan.
       avg_sa_to_vol (float):
          The average ratio of surface area to volume for the cylinders in the collection. Defaults to np.nan.
       max_branch_order (float):
          The maximum branch order in the collection. Defaults to np.nan.
       max_reverse_branch_order (float):
          The maximum reverse branch order in the collection. Defaults to np.nan.
       extent (dict):
          The minimum and maximum coordinates of the cylinders in the collection.
          Defaults to {
               "min": [np.nan, np.nan, np.nan],
               "max": [np.nan, np.nan, np.nan],
           }.
       treeQualities (dict):
          Various characteristics of the tree represented by the collection. Defaults to {
           "total_psa": None,
           "tot_hull_area": None,
           "stem_flow_hull_area": None,
           "stem_psa": None,
           "flowStats": None,
           "dbh": None,
           "tot_surface_area": None,
           "stem_surface_area": None,
       }.
       projections (dict):
          Whether the cylinders have been projected onto the XY, XZ, and YZ planes.
          Defaults to {
           "XY": False,
           "XZ": False,
           "YZ": False
           }.
       hulls (dict):
          Whether hulls have been calculated for the projections onto the XY, XZ, and YZ planes.
          Defaults to {
           "XY":False,
           "XZ": False,
           "YZ": False
          }.
       stem_hull (shapely.Polygon):
          The hull of the stem of the tree. Defaults to None.
       stem_paths (list):
          The paths along the stem of the tree. Defaults to [[]].
       trunk (list):
          The cylinders that make up the trunk of the tree. Defaults to [].
       graph (nx.Graph):
          The graph representation of the tree. Defaults to None.
       digraph (nx.Digraph):
          The directed graph representation of the tree. Defaults to None.
       end_nodes (list):
          The nodes at the ends of the branches of the tree. Defaults to [].
       flows (list[Flow]):
          The flows of water intercepted by the cylinders. Defaults to None.
       divide_nodes (list[int]):
          The nodes at which the flows divide. Defaults to None.
       drip_nodes (list[int]):
          The nodes at which water drips from the tree. Defaults to None.
       cyl_to_drip (dict[int,int]):
          A mapping from cylinders to drip points. Defaults to None.
       drip_points (dict[str,list]):
          The coordinates and flow ids of the drip points.
          Defaults to {
           "x": np.nan,
           "y": np.nan,
           "z": np.nan,
           "flow_id": np.nan
           }.
       trunk_nodes (list):
          The nodes that make up the trunk of the tree. Defaults to [].
       drip_point_loc (list):
          The locations of the drip points. Defaults to None.
       stem_flow_component (nx.Graph):
          The component of the graph representing stem flow. Defaults to None.
       drip_graph (nx.Graph):
          The graph representing the flow of drips. Defaults to None.
       drip_flow_components (list[nx.Graph]):
          The components of the drip graph. Defaults to None.
       flow_chars (dict):
          Characteristics of the flows: num_cylinders, projected_area,
           surface_area, angle_sum, volume, sa_to_vol, drip_node_id, drip_node_loc.
          Defaults to {}.
       trunk_lean (float):
          The angle of lean of the trunk. Defaults to None.
    """

    cylinders = dict

    # initialize our object level variables for cylinder objects
    def __init__(self) -> None:
        self.file = ""
        # Aggregate values from file
        self.surface_area = np.nan
        self.file_name = ""
        self.pSV = []  # Whole tree polygon
        self.volume = np.nan
        self.avg_sa_to_vol = np.nan
        self.max_branch_order = np.nan
        self.max_reverse_branch_order = np.nan
        self.extent = {
            "min": [np.nan, np.nan, np.nan],
            "max": [np.nan, np.nan, np.nan],
        }
        # to populate with x,y,z mins and maxs
        self.treeQualities = {
            "total_psa": None,
            "tot_hull_area": None,
            "stem_flow_hull_area": None,
            "stem_psa": None,
            "flowStats": None,
            "dbh": None,
            "tot_surface_area": None,
            "stem_surface_area": None,
        }

        # Projection Attrs
        self.projections = {"XY": False, "XZ": False, "YZ": False}
        self.hulls = {"XY": False, "XZ": False, "YZ": False}
        self.stem_hull = None

        # Special case tree attributes
        self.stem_paths = [[]]
        self.trunk = []

        # Graph and Attributes
        self.graph = None
        self.digraph = None
        self.end_nodes = []
        self.flows = None
        self.divide_nodes = None
        self.drip_nodes = None
        self.cyl_to_drip = None
        self.drip_points = {"x": np.nan, "y": np.nan, "z": np.nan, "flow_id": np.nan}
        self.trunk_nodes = []
        self.drip_point_loc = None
        self.stem_flow_component = None
        self.drip_graph = None
        self.drip_flow_components = None
        # Calculations using graph results
        self.flow_chars = {}
        self.trunk_lean = None

    def __eq__(self, other):
        """
        Compares two CylinderCollections for equality.

        Args:
            other: The other CylinderCollection to compare.

        Returns:
            bool: True if the cylinders in the two collections are the same, False otherwise.

        Raises:
            AttributeError: If one or both CylinderCollections contain no cylinders.
            TypeError: If the other object is not a CylinderCollection.

        """
        if len(self.cylinders) == 0 or len(other.cylinders) == 0:
            raise AttributeError(
                "One or both Cylinder Collections contain no Cylinders (Did you forget to initialize?)"
            )
        if type(self) == type(other):
            raise TypeError(
                "CylinderCollections may only be compared to other cylinder collections"
            )
        # order matters here
        return np.all([cyl == other[idx] for idx, cyl in enumerate(self.cylinders)])

    def from_csv(self, file: TextIOWrapper | str) -> None:
        """
        Initializes a new Cylinder Collection based on the QSM data.
         the csv file is read in with the columns mapped to the configured column locations.

        Args:
            file: The file object or file name to read the data from.

        """
        if isinstance(file, str):
            self.file_name = file
            file = open(input_dir + file)
        else:
            self.file = file
            self.file_name = file.name
        log.info(f"Processing {str(file)}")
        self.arr = np.genfromtxt(file, delimiter=",", skip_header=True)[0:, :-1]
        cylinders = [create_cyl(row) for row in self.arr]
        self.cylinders = cylinders

        min_x = np.min([cyl.x[0] for cyl in cylinders])
        min_y = np.min([cyl.y[0] for cyl in cylinders])
        min_z = np.min([cyl.z[0] for cyl in cylinders])
        max_x = np.max([cyl.x[1] for cyl in cylinders])
        max_y = np.max([cyl.y[1] for cyl in cylinders])
        max_z = np.max([cyl.z[1] for cyl in cylinders])
        # Aggregate values from file
        self.no_cylinders = len(cylinders)
        self.surface_area = np.sum([cyl.surface_area for cyl in cylinders])
        self.volume = np.sum([cyl.volume for cyl in cylinders])
        self.max_branch_order = np.max([cyl.branch_order for cyl in cylinders])
        self.max_reverse_branch_order = np.max(
            [cyl.reverse_branch_order for cyl in cylinders]
        )
        self.avg_sa_to_vol = (
            np.sum([cyl.sa_to_vol for cyl in cylinders]) / self.no_cylinders
        )
        self.extent = {
            "min": [min_x, min_y, min_z],
            "max": [max_x, max_y, max_z],
        }

        self.theta = np.nan
        log.info(f"{file.name} initialized with {self.no_cylinders} cylinders")

    def project_cylinders(self, plane: str = "XY") -> None:
        """
        Projects cylinders onto the specified plane.

        Args:
            plane (str): The plane to project the cylinders onto. Valid values are "XY", "XZ", and "YZ".

        """
        if plane not in ("XY", "XZ", "YZ"):
            log.info(f"{plane}: invalid value for plane")

        else:
            polys = []
            log.info(f"Projection into {plane} axis begun for file {self.file_name}")
            for idx, cyl in enumerate(self.cylinders):
                poly = cyl.get_projection(plane)
                polys.append(poly)
                # print a progress update once every 10 thousand or so cylinders
                intermitent_log(idx, self.no_cylinders, "Cylinder projection: ")
            # Used by other functions to know what projections have been run
            self.projections[plane] = True
            self.pSV = polys
            log.info(f"Projection into {plane} axis complete for file {self.file_name}")

    def get_collection_data(self) -> list[str]:
        """
        Returns a list of string representations of the cylinders in the collection.

        Returns:
            list[str]: A list of string representations of the cylinders.
        """
        cyl_desc = [cyl.__repr__() for cyl in self.cylinders]
        return cyl_desc

    def draw(
        self,
        plane: str = "XY",  # XY, XZ, YZ, or 3D
        highlight_lambda: Callable = lambda: True,
        filter_lambda: Callable = lambda: True,
        include_drips: bool = False,
        include_alpha_shape: bool = False,
        overlay_transparency: float = 0.5,
        file_name_ext: str = "",
        **kwargs,
    ):
        """
        Draws cylinders meeting given characteristics onto the specified plane.

        Args:
            plane:
                The plane to draw the cylinders on. Valid values are "XY", "XZ", "YZ", and "3D".
            highlight_lambda:
                A lambda function that determines which cylinders to highlight.
                Takes no inputs but may reference any attribute of a cylinder object.
            filter_lambda:
                A lambda function that determines which cylinders to include in the drawing.
                Takes no inputs but may reference any attribute of a cylinder object.
            include_drips:
                If True, includes the most recently set drip points in the drawing.
                Default is False.
            include_alpha_shape:
                If True, includes the most recently created alpha shape in the drawing.
                Default is False.
            overlay_transparency:
                Determines the transparency of the drip points and alpha shape drawn.
                Default is 0.5,
            file_name_ext:
                A suffix and/or the file extension to use for the output file. Default is an empty string.
            **kwargs:
            Additional keyword arguments to pass to the drawing function.
        """
        if plane not in ("XY", "XZ", "YZ", "3D"):
            log.info(f"{plane}: invalid value for plane")
        file_name = f"{self.file_name}_{plane}_{file_name_ext}"
        cylinders, _ = lam_filter(self.cylinders, filter_lambda)
        filtered_cyls, matches = lam_filter(
            cylinders, highlight_lambda, return_all=True
        )
        log.info(f"{len(filtered_cyls)} cylinders matched criteria")
        if plane == "3D":
            radii, vectors = [], []
            for cyl in filtered_cyls:
                start_end = np.array(
                    [
                        np.array([cyl.x[0], cyl.y[0], cyl.z[0]]),
                        np.array([cyl.x[1], cyl.y[1], cyl.z[1]]),
                    ]
                )
                radius = cyl.radius
                radii.append(radius)
                vectors.append(start_end)
            fig = draw_cylinders_3D(radii, vectors, file_name_ext=file_name, **kwargs)
        else:
            if not self.projections[plane]:
                self.project_cylinders(plane)
            to_draw = [cyl.projected_data[plane]["polygon"] for cyl in filtered_cyls]

            overlay = []
            if include_alpha_shape:
                if not self.hulls[plane]:
                    log.warning(
                        f"No {plane} plane alpha shape found. Running watershed_boundary"
                    )
                    self.watershed_boundary(plane=plane)
                overlay.append(self.hulls[plane])

            if include_drips:
                if not self.drip_point_loc:
                    log.warning(
                        "No drip point locations found, running set_drip_points"
                    )
                    self.set_drip_points()
                dpts = self.drip_point_loc
                x_pts = list(pt[0] for pt in dpts)
                y_pts = list(pt[1] for pt in dpts)
                z_pts = list(pt[2] for pt in dpts)
                if plane == "XY":
                    overlay.append([x_pts, y_pts])
                elif plane == "XZ":
                    overlay.append([x_pts, z_pts])
                elif plane == "YZ":
                    overlay.append([y_pts, z_pts])

            fig = draw_cyls(
                collection=to_draw,
                colors=matches,
                file_name_ext=file_name,
                overlay=overlay,
                overlay_transparency=overlay_transparency,
                **kwargs,
            )
        return fig

    def get_dbh(self) -> float:
        """
        Calculate and return the diameter at breast height (DBH) of the tree.

        Returns:
            float: The diameter at breast height (DBH) of the tree.
        """
        if self.treeQualities["dbh"]:
            return self.treeQualities["dbh"]
        else:
            desired_height = 1.3
            approx_height_range_start = desired_height - 0.1
            start_z = self.extent["min"][2]

            diff_from_desired_height = [
                (abs(cyl.z[0] - desired_height), cyl.cyl_id)
                for cyl in self.cylinders
                if cyl.z[0] >= start_z + approx_height_range_start
                and cyl.branch_order == 0
            ]
            closest_to_target_id = np.argmin(np.array(diff_from_desired_height[:][0]))
            rbh = self.cylinders[
                int(diff_from_desired_height[closest_to_target_id][1])
            ].radius
            self.treeQualities["dbh"] = 2 * rbh
        return self.treeQualities["dbh"]

    def get_end_nodes(self) -> list[int]:
        """
        Returns a list of end nodes in the graph.

        If the end nodes have already been computed, it returns the saved result.
        Otherwise, it computes the end nodes by finding nodes with a degree of 1 in the graph.

        Returns:
            A list of end nodes in the graph.
        """
        if not self.digraph:
            log.warning(
                "Graph not initialized, running initialize_graph_from with default values"
            )
            self.initialize_digraph_from()
        g = self.digraph

        if self.end_nodes:
            return self.end_nodes
        elif len(g.nodes) > 0:
            # the root node also has degree one, so we exclude it with n!= -1
            end_nodes = [n for n in g.nodes if g.degree(n) == 1 and n != -1]
            self.end_nodes = end_nodes
            return end_nodes

    def get_trunk_nodes(self) -> list[int]:
        """
        Returns a list of trunk nodes in the digraph.

        If the trunk nodes have already been computed, it returns the saved result.
        Otherwise, it computes the trunk nodes based on the branch order of the cylinders.

        Returns:
            A list of node ids for trunk nodes in the digraph.
        """
        g = self.digraph
        if self.trunk_nodes:
            return self.trunk_nodes
        elif len(g.nodes) > 0:
            trunk_cyls, _ = lam_filter(self.cylinders, lambda: branch_order == 0)  # noqa
            trunk_nodes = [cyl.cyl_id for cyl in trunk_cyls]
            self.trunk_nodes = trunk_nodes
            return trunk_nodes
        else:
            log.warning(
                "Graph not initialized, run <CylinderCollection>.initialize_graph(**args)"
            )
            return list(None)

    def direction_from_trunk(self, cyl: Cylinder):
        """
            Returns an array representing the direction of the cylinder from the trunk.

        Args:
            cyl: the cylinder for which to determine the direction from the trunk.

        Returns:
            tuple:
                A length 3 tuple corresponding the the x, y and z location of the cylinder.
                For each, a value of -1 indicates the cylinders location on the given axis
                    is less than the trunk's location on the same axis.
        """
        base_cyl = self.cylinders[0]
        tree_center = (base_cyl.x[0], base_cyl.y[0], base_cyl.z[0])
        x, y, z = cyl.x[1], cyl.y[1], cyl.z[1]
        directon_from_trunk = (
            -1 if x <= tree_center[0] else 1,
            -1 if y <= tree_center[1] else 1,
            -1 if z <= tree_center[1] else 1,
        )
        return directon_from_trunk

    def watershed_boundary(
        self,
        plane: str = "XY",
        filter_lambda: Callable | None = None,
        curvature_alpha: np.float64 = 1.8,
        draw: bool = False,
        save: bool = False,
        file_name_ext: str = "",
    ) -> None:
        """
        Generates tightly fit concave_hull (alpha shape) around the passed component.

        Args:
            component: The component for which to generate the watershed boundary. If None, the entire digraph is used.
            plane: The plane in which to generate the watershed boundary. Defaults to "XY".
            filter_lambda:
                A lambda function that determines which cylinders to include in the drawing.
                Takes no inputs but may reference any attribute of a cylinder object.
            curvature_alpha: The alpha value that determines the tightness of the fit of the shape.
                             Too low an alpha results in multiple shapes. Defaults to 2.
            draw: If True, the hull polygon is drawn via matplotlib. Defaults to False.
            save: If True, the hull is drawn and saved. Defaults to False.
            file_name_ext:
                The file extension to add to the end of the file name when when saving the drawn hull. Defaults to an empty string.

        Returns:
            None
        """
        comp_cylinders = self.cylinders
        if filter_lambda:
            comp_cylinders, _ = lam_filter(self.cylinders, filter_lambda)

        end_nodes = self.get_end_nodes()
        end_cylinders = [
            cyl
            for cyl in comp_cylinders
            if cyl.cyl_id in end_nodes or cyl.branch_order == 0
        ]
        boundary_points = [
            (cyl.x[1], cyl.y[1], cyl.z[1] + cyl.radius * 2) for cyl in comp_cylinders
        ]
        if plane in ("YZ", "XZ"):
            # using only branch tips as boundary points only really
            #  works in the XY direction. In the other directions
            #  the botom of the canopy is bounded by points on the surfaces
            # of various cylinders
            comp_cylinders = [cyl for cyl in comp_cylinders if cyl.branch_order != 0]
            for cyl in comp_cylinders:
                buffer = cyl.radius * 2
                buffer_arr = np.array(self.direction_from_trunk(cyl)) * buffer
                coords = (cyl.x[0], cyl.y[0], cyl.z[0])
                buffered_coords = coords + buffer_arr
                boundary_points.append(buffered_coords)

        if plane == "XY":
            boundary_points = [Point(cyl.x[1], cyl.y[1]) for cyl in end_cylinders]
        elif plane == "XZ":
            boundary_points = [Point(x, z) for x, _, z in boundary_points]
        elif plane == "YZ":
            boundary_points = [Point(y, z) for _, y, z in boundary_points]
        hull, _, _ = concave_hull(boundary_points, curvature_alpha)
        if draw:
            draw_cyls([hull], save=save, file_name_ext=file_name_ext)
        self.hulls[plane] = hull
        return hull, boundary_points

    def initialize_digraph_from(self, in_flow_grade_lim=in_flow_grade_lim):
        """This function creates a directed graph and its undirected counterpart.
        Initializes edge attributes as cylinder objects

        Args:
            in_flow_grade_lim: The drip edge cut off angle.
            Defaults to the in_flow_grade_lim from the configuration file.

        """
        gr = nx.DiGraph()
        trunk_nodes, _ = lam_filter(self.cylinders, lambda: branch_order == 0)  # noqa
        edges = (
            (
                (int(cyl.cyl_id), int(cyl.parent_id), {"cylinder": cyl})
                if (cyl.angle >= in_flow_grade_lim or cyl.cyl_id in trunk_nodes)
                else (int(cyl.parent_id), int(cyl.cyl_id), {"cylinder": cyl})
            )
            for cyl in self.cylinders
        )
        gr.add_edges_from(edges)
        self.digraph = gr
        self.graph = gr.to_directed()

    def contracted_nodes(self, node_list, next_node, G) -> tuple:
        """
        Returns a graph made from removing nodes from G,
        and connecting the removed nodes' neighbors via an edge

        Args:
            node_list:
                Nodes to be removed from the graph and connected to the next node.
            next_node:
                The node to which the removed nodes should be connected.
            G:
                The graph on which to carry out the operation

        Returns:
            nx.Graph: A graph with the listed nodes removed and their neighbors connected to the next node.
            list: A list of the neighbors of the removed nodes
        """
        # finding and removing the listed nodes to find their edges
        contracted = copy.deepcopy(G)
        contracted.remove_nodes_from(node_list)

        neighbors = [
            node
            for node in contracted.nodes()
            if (G.degree(node) != contracted.degree(node))
        ]

        contracted.add_node(next_node)
        edges_to_trunk = [(u, next_node) for u in neighbors]
        contracted.add_edges_from(edges_to_trunk)

        return contracted, neighbors

    def find_trunk_distances(self) -> dict:
        """
        Finds the distance in the graph (in number of nodes) between each node and the closest trunk node
        """
        trunk_nodes = self.get_trunk_nodes()

        trunk_contraction, _ = self.contracted_nodes(trunk_nodes, 0, self.graph)
        trunk_paths = nx.shortest_path(trunk_contraction, target=0)
        dists = {node: len(path) - 1 for node, path in trunk_paths.items()}
        return dists

    def find_flow_components(self, save_components: bool | None = False):
        """
        Determines to which partition perciplitation intercepted by
            each cylinder in the collection is likely to contribute to.
        That is, viewing each cylinder as a source, and allowing travel on the
            digraph as defined by the edge's direction, the function determines:
            - the set of nodes that can be reached from each source node.
            - for each set of nodes (each, 'flow') the eventual sink (or drip nodes)
                to which the theoretical flow is directed

        Args:
            save_components:
                If true, the stemflow and drip components of the collection's graph
                are saved. This is rather memory intensive so this input defaults to False.

        Raises:
            TypeError: Raised if the function is called on an undirected graph.
        """
        g = self.digraph
        if type(g) == nx.Graph:
            msg = "Find Flow Digraph invoked for undirected graph"
            log.error(msg)
            raise TypeError(msg)
        root_node = -1

        divide_nodes = [
            node
            for node in g.nodes()
            if g.out_degree(node) > 1
            or (g.out_degree(node) == 1 and g.in_degree(node) == 0)
        ]
        drip_nodes = [
            node
            for node, out_degree in g.out_degree()
            if out_degree == 0 and node != -1
        ]
        stem_flow_component = g.subgraph(nx.ancestors(g, root_node) | {0}).copy()

        stem_cylinders = [
            node
            for node in stem_flow_component.nodes()
            if stem_flow_component.degree(node) > 0
        ]
        for cyl in self.cylinders:
            if cyl.cyl_id in stem_cylinders:
                cyl.is_stem = True

        g_drip = copy.deepcopy(g)
        g_drip.remove_edges_from(stem_flow_component.edges())
        # g_drip.remove_nodes_from(stem_flow_component.nodes())

        drip_divide_pairings = [
            (
                drip_node,
                [node for node in divide_nodes if nx.has_path(g_drip, node, drip_node)],
            )
            for drip_node in drip_nodes
            if drip_node != -1
        ]

        drip_components = []
        cyl_to_drip_node = {cyl.cyl_id: [] for cyl in self.cylinders}
        for pair in drip_divide_pairings:
            drip_node, source_divides = pair
            paths = [
                nx.shortest_path(g_drip, div_node, drip_node)
                for div_node in source_divides
            ]
            component_nodes = chain.from_iterable(paths)
            component_graph = g_drip.subgraph(component_nodes).copy()
            component_cyls = [
                attr["cylinder"].cyl_id
                for _, _, attr in component_graph.edges(data=True)
            ]
            for idx in component_cyls:
                cyl_to_drip_node[idx].append(drip_node)

            drip_components.append((drip_node, g_drip.subgraph(component_nodes).copy()))

        for cyl in self.cylinders:
            if cyl_to_drip_node[cyl.cyl_id]:
                cyl.drip_node = cyl_to_drip_node[cyl.cyl_id][-1]

        log.info(
            f"{self.file_name} found to have {len(drip_components)} drip components"
        )
        print("reached_End of find flows")
        self.divide_nodes = divide_nodes
        self.drip_nodes = drip_nodes
        self.cyl_to_drip = cyl_to_drip_node
        if save_components:
            self.stem_flow_component = stem_flow_component
            self.drip_flow_components = drip_components
            self.drip_graph = g_drip

    # @profile
    def calculate_flows(self, plane: str = "XY"):
        """
        Based on the findings of 'find_flows', sums the qualities of
            all of the cylinders in each flow and stores the results in
            flow objects.
        Args:
            plane:
                The projection for which to source the angle and projected area
                data sumed. Valid values are "XY", "XZ", "YZ", and "3D".
        """
        if not self.projections[plane]:
            log.warning(
                f"""No {plane} projection data found. Running project_cylinders
                            This may take several minutes"""
            )
            self.project_cylinders(plane)
        cyls = self.cylinders
        flow_chars = []
        # edge_attributes = {}

        # log.info("attempting to sum stem edges ")

        # this probably doesn't belong here but its efficient to do it now
        # is 'needed' for statistics section
        np_flow_chars = [None] * (len(self.drip_nodes) + 1)

        def numpy_flow_chars(lambda_filter: Callable, drip_cyl, index: int):
            arr = np.array(
                [
                    np.array(
                        [
                            1,
                            np.float64(cyl.projected_data[plane]["area"]),
                            cyl.surface_area,
                            cyl.angle,
                            cyl.volume,
                            cyl.sa_to_vol,
                        ]
                    )
                    for cyl in cyls
                    if lambda_filter(cyl)
                ]
            )
            flow = np.sum(arr, axis=0)
            np_flow_chars[index] = Flow(
                flow[0],
                flow[1],
                flow[2],
                flow[3],
                flow[4],
                flow[5],
                drip_cyl.cyl_id,
                (drip_cyl.x[0], drip_cyl.y[0], drip_cyl.z[0]),
            )

        numpy_flow_chars(
            lambda_filter=lambda x: x.is_stem, drip_cyl=self.cylinders[0], index=0
        )
        log.info(f"np flow chars {np_flow_chars}")
        log.info(f" flow chars {flow_chars}")
        flow_chars.append(np_flow_chars[0])
        log.info(f" togetha {flow_chars}")
        # log.info(f"summed stem edges {flow_chars}")
        for idx, drip_node in enumerate(self.drip_nodes):
            cyl_before_drip = [cyl for cyl in cyls if cyl.cyl_id == drip_node]

            if len(cyl_before_drip) > 1:
                log.warning(f"Error: More that 1 cyl with id {drip_node} found")

            drip_node_loc = (
                cyl_before_drip[0].x[1],
                cyl_before_drip[0].y[1],
                cyl_before_drip[0].z[1],
            )
            drip_cylinders = [
                cyl
                for cyl in cyls
                if cyl.drip_node == drip_node and cyl.cyl_id != drip_node
            ]

            flow_chars.append(
                Flow(
                    **{
                        "num_cylinders": len(drip_cylinders),
                        "projected_area": np.sum(
                            [
                                np.float64(cyl.projected_data[plane]["area"])
                                for cyl in drip_cylinders
                            ]
                        ),
                        "surface_area": np.sum(
                            [cyl.surface_area for cyl in drip_cylinders]
                        ),
                        "angle_sum": np.sum([cyl.angle for cyl in drip_cylinders]),
                        "volume": np.sum([cyl.volume for cyl in drip_cylinders]),
                        "sa_to_vol": np.sum([cyl.sa_to_vol for cyl in drip_cylinders]),
                        "drip_node_id": drip_node,
                        "drip_node_loc": drip_node_loc,
                    }
                )
            )
        self.flows = flow_chars

    def identify_stem_paths(self):
        """
        Identifies the paths in the collection's graph by which the root node can be reached.
        """
        # Special case tree attributes
        root = 0
        self.stem_paths = nx.all_simple_paths(self.g, 0, root)

    def find_furthest(node, nodes):
        """Finds the furthest node in the given list from the provided n"""
        nodes = np.asarray(nodes)
        dist_2 = np.sum((nodes - node) ** 2, axis=1)
        return np.argmax(dist_2)

    def find_trunk_lean(self):
        """
        Defines a straight line from base to tip of the trunk then
         finds the angle of that line from the XZ plane
        """
        trunk_beg = lam_filter(self.cylinders, a_lambda=lambda: branch_order == 0)  # noqa
        trunk_points = [(cyl.x[0], cyl.y[0], cyl.z[0]) for cyl in trunk_beg]
        root = trunk_points[0]
        furthest_afeild, _ = furthest_point(root, trunk_points)
        dx, dy, dz = ((a - b) for a, b in zip(furthest_afeild, root))
        run = np.sqrt(dx**2 + dy**2)
        angle = np.arctan(dz / np.sqrt(dx**2 + dy**2)) if run > 0 else np.arctan(0)
        self.trunk_lean = angle
        return angle

    def find_overlap_by_percentile(
        self, plane: str = "XY", percentiles: list[int] = [25, 50, 75]
    ):
        """
        Groups the projected polygons of the collection's cylinders
            by percentiles given - with respect to their z values.
        Passes the list of grouped polygons to the get_projected_overlap function
            to get the overlap percentages between the groups.
        See get_projected_overlap for more information.

        Args:
            plane:
                Plane of projection from which to source the polygons.
                Defaults to "XY".
            percentiles:
                Percentiles defining the groups passed to get_projected_overlap.
                Expected to be strictly increasing. Defaults to [25, 50, 75].

        Returns:
            list[dict]: Each dicitonary corresponding to a percentile,
            with metrics describing the overlap between the polygons in that percentile
            {
                "sum_area": sum of areas of the polygons in the percentile,
                "effective_area": the area of the union of the polygons in the percentile,
                "internal_overlap": sum_area - effective_area,
                "overlap_with_previous": the area of the overlap between the polygons in the percentile,
                        with the polygons in the previous percentile,
            }
        """
        percentiles.sort()
        if not self.projections[plane]:
            self.project_cylinders(plane)
        non_trunk_polys, _ = lam_filter(self.cylinders, lambda: branch_order != 0)  # noqa
        cyl_metric = [cyl.z[0] for cyl in non_trunk_polys]
        percentiles_by_metric = np.percentile(cyl_metric, percentiles)
        prev_perc = 0
        poly_matrix = []
        for perc in percentiles_by_metric:
            poly_matrix.append(
                [
                    cyl.projected_data[plane]["polygon"]
                    for cyl in non_trunk_polys
                    if cyl.z[0] <= perc and cyl.z[0] > prev_perc
                ]
            )
            prev_perc = perc
        overlaps = get_projected_overlap(poly_matrix, percentiles)

        return overlaps

    def statistics(self, plane: str = "XY", file_name_ext: str = ""):
        """Generates a csv and dictionary containing a variety of descriptive
            statistics about the collection of cylinders.

        Args:
            plane:
                The plane used for projection dependent statistics.
                Defaults to "XY".
            file_name_ext:
                the naming suffix to used for the saved file name

        Returns:
            _description_
        """
        if not self.pSV:
            self.project_cylinders(plane)
        if not self.hulls[plane]:
            self.watershed_boundary(plane=plane)
        if not self.stem_hull:
            if not self.stem_flow_component:
                self.find_flow_components()
                self.calculate_flows(plane=plane)
            self.stem_hull, _ = self.watershed_boundary(
                plane="XY",
                filter_lambda=lambda: is_stem,  # noqa
            )
        dbh = self.get_dbh()

        canopy_cover = self.hulls[plane].area
        canopy_boundary = self.hulls[plane].boundary.length

        canopy_cover_stem = self.stem_hull.area
        canopy_boundary_stem = self.stem_hull.boundary.length

        log.info("Found hull alpha shape stats")

        # calculate projected areas and (therefore) overlaps
        # Unary union gives a single contiguous polygon when fed many overlapping cylinders
        # The area of the union thus differs from the sum of the areas of its components
        # in that the former counts overlaps only once
        # polys = [poly_dict['polygon'] for poly_dict in self.pSV]
        polys = self.pSV
        tot_poly = unary_union(polys)
        projected_union_area = tot_poly.area
        sum_projected_area = np.sum([poly.area for poly in self.pSV])
        stem_polys = [
            cyl.projected_data[plane]["polygon"]
            for cyl in self.cylinders
            if cyl.is_stem
        ]
        union_poly_stem = unary_union(stem_polys)
        projected_union_area_stem = union_poly_stem.area
        sum_projected_area_stem = np.sum([poly.area for poly in stem_polys])

        log.info("found projected areas")

        # this could techically be conmbined with the above by adding a percentile of 100
        #
        overlap_dict = self.find_overlap_by_percentile(percentiles=[25, 50, 75])

        tot_poly = unary_union(self.pSV)

        min_x = self.extent["min"][0]
        max_x = self.extent["max"][0]
        min_y = self.extent["min"][1]
        max_y = self.extent["max"][1]
        min_z = self.extent["min"][2]
        max_z = self.extent["max"][2]

        if self.flows:
            stem_flow = [flow for flow in self.flows if flow.drip_node_id == 0][0]
        else:
            stem_flow = 0

        total_surface_area = np.sum([cyl.surface_area for cyl in self.cylinders])
        total_volume = np.sum([cyl.volume for cyl in self.cylinders])
        max_bo = np.max([cyl.branch_order for cyl in self.cylinders])

        order_zero_cyls, _ = lam_filter(self.cylinders, lambda: branch_order == 0)  # noqa
        order_one_cyls, _ = lam_filter(self.cylinders, lambda: branch_order == 1)  # noqa
        order_two_cyls, _ = lam_filter(self.cylinders, lambda: branch_order == 2)  # noqa
        order_three_cyls, _ = lam_filter(self.cylinders, lambda: branch_order == 3)  # noqa
        order_four_cyls, _ = lam_filter(self.cylinders, lambda: branch_order == 4)  # noqa

        statistics = {
            "total_psa": projected_union_area,
            "psa_w_overlap": sum_projected_area,
            "stem_psa": projected_union_area_stem,
            "stem_psa_w_overlap": sum_projected_area_stem,
            "tot_surface_area": total_surface_area,
            "stem_surface_area": stem_flow.surface_area,
            "tot_hull_area": canopy_cover,
            "tot_hull_boundary": canopy_boundary,
            "stem_hull_area": canopy_cover_stem,
            "stem_hull_boundary": canopy_boundary_stem,
            "num_drip_points": len(self.drip_nodes),
            "max_bo": max_bo,
            "topQuarterTotPsa": overlap_dict[75]["sum_area"],
            "topHalfTotPsa": overlap_dict[50]["sum_area"],
            "topThreeQuarterTotPsa": overlap_dict[25]["sum_area"],
            "TotalShade": sum_projected_area - projected_union_area,
            "top_quarter_shade": overlap_dict[75]["overlap_with_previous"],
            "top_half_shade": overlap_dict[50]["overlap_with_previous"],
            "top_three_quarter_shade": overlap_dict[25]["overlap_with_previous"],
            "DBH": dbh,
            "volume": total_volume,
            "X_max": max_x,
            "Y_max": max_y,
            "Z_max": max_z,
            "X_min": min_x,
            "Y_min": min_y,
            "Z_min": min_z,
            "Order_zero_angle_avg": np.average([cyl.angle for cyl in order_zero_cyls]),
            "Order_zero_angle_std": np.std([cyl.angle for cyl in order_zero_cyls]),
            "Order_one_angle_avg": np.average([cyl.angle for cyl in order_one_cyls]),
            "Order_one_angle_std": np.std([cyl.angle for cyl in order_one_cyls]),
            "Order_two_angle_avg": np.average([cyl.angle for cyl in order_two_cyls]),
            "Order_two_angle_std": np.std([cyl.angle for cyl in order_two_cyls]),
            "Order_three_angle_avg": np.average(
                [cyl.angle for cyl in order_three_cyls]
            ),
            "Order_three_angle_std": np.std([cyl.angle for cyl in order_three_cyls]),
            "order_gr_four_angle_avg": np.average(
                [cyl.angle for cyl in order_four_cyls]
            ),
            "order_gr_four_angle_std": np.std([cyl.angle for cyl in order_four_cyls]),
            "file_name": self.file_name + plane,
        }
        stat_file = save_file(
            self.file_name.replace(".csv", f"_{file_name_ext}"),
            out_file=statistics,
            method="statistics",
            overwrite=True,
        )

        return stat_file

    def generate_flow_file(self, file_name_ext):
        """Creates a csv and dictionary containing data describing
        the flows calculated in find_flow_components and calculate_flows

        Args:
            file_name_ext:
                the naming suffix to used for the saved file name

        Returns:
            dict: A dictionary containing the flow data
        """
        flow_dicts = [flow.__dict__ for flow in self.flows]
        flow_file = save_file(
            self.file_name.replace(".csv", f"_flows_{file_name_ext}"),
            out_file=flow_dicts,
            subdir="flows",
            method="flows",
            overwrite=True,
        )
        return flow_file

    def set_drip_points(
        self,
        # metric :str = 'projected_area',
        percentile: int = 90,
    ):
        """
        Returns the locations of the drip points.
        **Drip points are identified as the subset of drip nodes
            with a value for the provided metric in the percentile given
        """
        if not self.flows:
            log.warning(
                """flows are not defined, running 'find_flow_components'
                            and  'calculate flows'"""
            )
            log.warning("This may take several minutes")
            self.find_flow_components()
            self.calculate_flows()
        # excluding trunk node
        drip_nodes = [
            (f.drip_node_id, f.projected_area, f.drip_node_loc)
            for f in self.flows
            if f.drip_node_id != 0
        ]

        distinct_drip_node_ids = {node[0] for node in drip_nodes}

        flow_by_area_dict = {}
        for node in distinct_drip_node_ids:
            area = np.sum([flow[1] for flow in drip_nodes if flow[0] == node])
            flow_by_area_dict[node] = area

        point_cutoff = np.percentile(
            [area for _, area in flow_by_area_dict.items()], percentile
        )
        drip_points = [
            node for node, area in flow_by_area_dict.items() if area > point_cutoff
        ]
        if drip_points == []:
            log.warning(f"No drip points found in the {percentile}th percentile")
            log.warning("Assuming a small drip point list, including all drip points")
            drip_points = [node for node, _ in flow_by_area_dict.items()]
        drip_point_locs = list(
            [node[2][0], node[2][1], node[2][2]]
            for node in drip_nodes
            if node[0] in drip_points
        )
        self.drip_point_loc = drip_point_locs
        return drip_point_locs



"""This is a sample python file for testing functions from the source code."""

from __future__ import annotations

import os
import sys
from test.expected_results import (drip_adj_flows, drip_adj_stem_map,
                                   drip_mid_flows, drip_mid_stem_map,
                                   drip_on_trunk_flows, drip_on_trunk_stem_map,
                                   ez_projection_cyls, ez_projection_xy,
                                   ez_projection_xz, ez_projection_yz,
                                   happy_path_cyls, happy_path_dbh,
                                   happy_path_edges, happy_path_flows,
                                   happy_path_is_stem, small_tree_dbh,
                                   small_tree_edges, small_tree_flows,
                                   small_tree_is_stem, ten_cyls_bo_and_len,
                                   ten_cyls_bo_one, ten_cyls_cyls,
                                   ten_cyls_dbh, ten_cyls_edges,
                                   ten_cyls_flows, ten_cyls_id_one,
                                   ten_cyls_is_stem)
from test.expected_results_shapes import (small_tree_overlap,
                                          small_tree_wateshed_poly)
from test.utils import within_range

import pytest

from canopyhydro.configuration import test_input_dir
from canopyhydro.CylinderCollection import (CylinderCollection,
                                            pickle_collection,
                                            unpickle_collection)
from canopyhydro.utils import lam_filter

sys.path.insert(0, os.path.dirname(os.getcwd()))

create_cylinders_cases = [
    # (file, expected_cylinders )
    pytest.param("1_TenCyls.csv", ten_cyls_cyls, id="Ten Cyls"),
    pytest.param("2_EZ_projection.csv", ez_projection_cyls, id="EZ Projection"),
    pytest.param("3_HappyPathWTrunk.csv", happy_path_cyls, id="Happy Path"),
]

ez_projection_cases = [
    # (file, angles, projection )
    pytest.param("2_EZ_projection.csv", ez_projection_xy, "XY", id="XY projection"),
    pytest.param("2_EZ_projection.csv", ez_projection_xz, "XZ", id="XZ projection"),
    pytest.param("2_EZ_projection.csv", ez_projection_yz, "YZ", id="YZ projection"),
]

lam_filter_cases = [
    # (file, angles, projection )
    pytest.param(
        "1_TenCyls.csv",
        ten_cyls_bo_one,
        lambda: branch_order == 1,  # noqa
        id="XY projection",
    ),
    pytest.param(
        "1_TenCyls.csv",
        ten_cyls_bo_and_len,
        lambda: branch_order == 0 or length <= 0.22447,  # noqa
        id="XZ projection",
    ),
    pytest.param(
        "1_TenCyls.csv",
        ten_cyls_id_one,
        lambda: cyl_id == 1,  # noqa
        id="YZ projection",
    ),
]

find_flows_cases = [
    # (file, expected_stem_map, expected_flows )
    pytest.param("1_TenCyls.csv", ten_cyls_is_stem, ten_cyls_flows, id="Ten Cyls"),
    pytest.param(
        "7_DripPathAdjToTrunk.csv",
        drip_adj_stem_map,
        drip_adj_flows,
        id="Drip Adjacent Trunk",
    ),
    pytest.param(
        "8_DripPathMidBranch.csv",
        drip_mid_stem_map,
        drip_mid_flows,
        id="Drip Mid Branch",
    ),
    pytest.param(
        "9_DripOnTrunk.csv",
        drip_on_trunk_stem_map,
        drip_on_trunk_flows,
        id="Drip On Trunk",
    ),
    pytest.param(
        "3_HappyPathWTrunk.csv", happy_path_is_stem, happy_path_flows, id="Happy Path"
    ),
    pytest.param(
        "5_SmallTree.csv", small_tree_is_stem, small_tree_flows, id="Small Tree"
    ),
]


pickle_cases = [
    # (file, expected_stem_map, expected_flows )
    pytest.param("5_SmallTree.csv", id="Small Tree"),
    pytest.param("7_DripPathAdjToTrunk.csv", id="Drip Adjacent Trunk"),
    pytest.param("8_DripPathMidBranch.csv", id="Drip Mid Branch"),
    pytest.param("9_DripOnTrunk.csv", id="Drip On Trunk"),
    pytest.param("3_HappyPathWTrunk.csv", id="Happy Path"),
]

create_graph_cases = [
    # (file, expected_edges )
    pytest.param("1_TenCyls.csv", ten_cyls_edges, id="Ten Cyls"),
    pytest.param("3_HappyPathWTrunk.csv", happy_path_edges, id="Happy Path"),
    pytest.param("5_SmallTree.csv", small_tree_edges, id="Small Tree"),
]

dbh_cases = [
    # (file, expected_edges )
    pytest.param("1_TenCyls.csv", ten_cyls_dbh, id="Ten Cyls (Too Short)"),
    pytest.param("3_HappyPathWTrunk.csv", happy_path_dbh, id="Happy Path"),
    pytest.param("5_SmallTree.csv", small_tree_dbh, id="Small Tree"),
]


@pytest.mark.parametrize(
    "basic_collection, expected_cylinders",
    create_cylinders_cases,
    indirect=["basic_collection"],
)
def test_create_cylinders(basic_collection, expected_cylinders):
    actual = basic_collection.get_collection_data()
    expected = expected_cylinders
    assert expected == actual


@pytest.mark.parametrize("file_name, expected_cylinders", create_cylinders_cases)
def test_create_cylinders_from_csv(file_name, expected_cylinders):
    csv_collection = CylinderCollection()
    file_path = "/".join([str(test_input_dir), file_name])
    file_obj = open(file_path)
    csv_collection.from_csv(file_obj, test_input_dir)
    actual = csv_collection.get_collection_data()
    assert actual == expected_cylinders


@pytest.mark.parametrize(
    "flexible_collection, angles, projection_axis",
    ez_projection_cases,
    indirect=["flexible_collection"],
)
def test_project_cylinders(
    flexible_collection, angles, projection_axis, accepted_err=0.03
):
    for idx, cyl in enumerate(flexible_collection.cylinders):
        expected_angle = angles[idx]
        actual_angle = cyl.projected_data[projection_axis]["angle"]
        if projection_axis == "XY":
            actual_setup_angle = cyl.angle
            assert within_range(expected_angle, actual_setup_angle, accepted_err)
        assert within_range(expected_angle, actual_angle, accepted_err)


@pytest.mark.parametrize(
    "basic_collection, expected_result, lam_func",
    lam_filter_cases,
    indirect=["basic_collection"],
)
def test_lam_filter(basic_collection, expected_result, lam_func):
    cyls_returned, _ = lam_filter(basic_collection.cylinders, lam_func)
    actual_result = [cyl.cyl_id for cyl in cyls_returned]
    assert actual_result == expected_result


@pytest.mark.parametrize(
    "basic_collection, expected_stem_map, expected_flows",
    find_flows_cases,
    indirect=["basic_collection"],
)
def test_find_flows(basic_collection, expected_stem_map, expected_flows):
    basic_collection.project_cylinders("XY")
    basic_collection.initialize_digraph_from()
    basic_collection.find_flow_components()
    basic_collection.calculate_flows()
    actual_flows = basic_collection.flows
    _, actual_stem_map = lam_filter(
        basic_collection.cylinders,
        lambda: is_stem,  # noqa
        return_all=True,
    )
    print(actual_flows)
    print(expected_flows)
    try:
        assert actual_flows == expected_flows
        assert actual_stem_map == expected_stem_map
    except AssertionError as e:
        # breakpoint()
        raise e


@pytest.mark.parametrize(
    "basic_collection, expected_stem_map, expected_flows",
    find_flows_cases,
    indirect=["basic_collection"],
)
def test_pickle(basic_collection, expected_stem_map, expected_flows):
    basic_collection.project_cylinders("XY")
    basic_collection.initialize_digraph_from()
    basic_collection.find_flow_components()
    basic_collection.calculate_flows()
    pickle_file = pickle_collection(basic_collection)

    unpickled_collection = unpickle_collection(pickle_file)

    actual_flows = unpickled_collection.flows
    _, actual_stem_map = lam_filter(
        unpickled_collection.cylinders,
        lambda: is_stem,  # noqa
        return_all=True,
    )
    print(actual_flows)
    print(expected_flows)
    assert actual_flows == expected_flows
    assert actual_stem_map == expected_stem_map


@pytest.mark.parametrize(
    "basic_collection, expected_dbh", dbh_cases, indirect=["basic_collection"]
)
def test_dbh(basic_collection, expected_dbh):
    basic_collection.get_dbh()
    actual_dbh = basic_collection.treeQualities["dbh"]
    assert actual_dbh == expected_dbh


@pytest.mark.parametrize("flexible_collection", ["5_SmallTree.csv"], indirect=True)
def test_collection_overlap(flexible_collection):
    flexible_collection.project_cylinders("XY")
    actual = flexible_collection.find_overlap_by_percentile()
    assert actual == small_tree_overlap


@pytest.mark.parametrize("flexible_collection", ["5_SmallTree.csv"], indirect=True)
def test_watershed(flexible_collection):
    flexible_collection.initialize_digraph_from()
    flexible_collection.project_cylinders("XY")
    flexible_collection.watershed_boundary(flexible_collection.graph)
    actual_poly = flexible_collection.hulls["XY"]
    expected_poly = small_tree_wateshed_poly
    assert actual_poly == expected_poly



from __future__ import annotations

from test.expected_results_shapes import (example_coords,
                                          funky_squares_overlap_areas,
                                          squares_projection_overlap,
                                          star_poly, start_poly_coords)
from test.utils import within_range

import pytest
from shapely.geometry import LineString, Point, Polygon

import src.canopyhydro.geometry as geometry


def test_2D_overlap():
    polys_bottom = [
        Polygon([(0, 0), (2, 0), (2, 2), (0, 2)]),
        Polygon([(1, 1), (3, 1), (3, 3), (1, 3)]),
    ]
    polys_mid = [
        Polygon([(2, 2), (4, 2), (4, 4), (2, 4)]),
        Polygon([(3, 3), (5, 3), (5, 5), (3, 5)]),
    ]
    polys_top = [Polygon([(4, 4), (6, 4), (6, 6), (4, 6)])]

    poly_matrix = [polys_bottom, polys_mid, polys_top]
    labels = ["bottom", "mid", "top"]
    actual_overlap_dict = geometry.get_projected_overlap(poly_matrix, labels)
    assert squares_projection_overlap == actual_overlap_dict


def test_proj_overlap():
    """
    Tests for diff number of  (outer list) and diff number of polys (outer list)
    """
    square1 = Polygon([(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)])
    square2 = Polygon([(0.25, 0.25), (0.25, 0.5), (0.5, 0.5), (0.5, 0.25)])
    square3 = Polygon([(1.0, 1.0), (1.0, 2.0), (2.0, 2.0), (2.0, 1.0)])
    square4 = Polygon([(1.5, 1.5), (1.5, 2.5), (2.5, 2.5), (2.5, 1.5)])
    bottom = [square1, square2]
    mid = [square1, square2, square3]
    top = [square1, square2, square3, square4]
    poly_matrix = [bottom, mid, top]
    actual = geometry.get_projected_overlap(poly_matrix, ["bottom", "mid", "top"])
    assert actual == funky_squares_overlap_areas


@pytest.mark.parametrize(
    "vector,magnitude,radius,e_angle,e_area",
    [
        ([(2, -1), (4, -1), (5, -1)], [-3, -5, -6], 1, -0.785398, 13.91),
        ([(1, -0.5), (1, -1.5), (1, -3)], [1.5, 2.5, 3], 1, 0.785398, 8.08),
        ([(-2, 1), (-4, 1), (-5, 1)], [3, 5, 6], 1, 0.785398, 13.91),
        ([(1, 4), (1, 6), (1, 7)], [3, 5, 6], 1, 0.785398, 13.91),
        ([(-1, -4), (-1, -6), (-1, -7)], [3, 5, 6], 1, 0.785398, 13.91),
        ([(1, -2), (1, -4), (1, -5)], [3, 5, 6], 1, 0.785398, 13.91),
    ],
)
def test_project_cyl(vector, magnitude, radius, e_angle, e_area):
    """
    Tests projection of cylinders parallel with:
    the XY plane, Z axis the line x=y (45 deg)
    """
    actual = geometry.numba_get_projection(vector, magnitude, radius)
    assert within_range(e_angle, actual["angle"], 0.02)
    assert within_range(e_area, actual["area"], 0.02)


@pytest.mark.parametrize(
    "points,expected",
    [
        ([(1, 1), (-1, 1), (1, -1)], [0.0, 0.5, 0.5]),
        ([(0.5, 0.5), (0, 0), (0.5, 0)], [0.5, 0.5, 0.0]),
        ([(1, 0), (0, 0), (0, 1)], [0.5, 0.0, 0.5]),
    ],
)
def test_lu_factor(points, expected):
    """
    tests circumcenter func with some mock examples
    (one, two, many points, size zero array, cooincindent coords)
    """
    actual = geometry.circumcenter_lapack(points)
    for idx, exp in enumerate(expected):
        assert within_range(exp, actual[idx], 0.02)


@pytest.mark.parametrize(
    "points",
    [
        ([(1, 1), (-1, 1), (1, -1)]),
        ([(0.5, 0.5), (0, 0), (0.5, 0)]),
        ([(1, 0), (0, 0), (0, 1)]),
    ],
)
def test_lapack_v_lu_factor(points):
    """
    Tests thse answers are the same so that we only really need
    to test one
    """
    actual = geometry.circumcenter_lu_factor(points)
    expected = geometry.circumcenter_lapack(points)
    for idx, exp in enumerate(expected):
        assert within_range(exp, actual[idx], 0.02)


@pytest.mark.parametrize(
    "points,expected",
    [
        ([(1, 1), (-1, 1), (1, -1)], 1.41),
        ([(5, 6), (4, 4), (5, 4)], 1.11),
        ([(0.5, 0.5), (0, 0), (0.5, 0)], 0.35),
    ],
)
def test_curcumradius(points, expected):
    """
    test with input center, and without
    corrd list of one two and many lengths (div by three and non div by three)
    test with 3+ dims
    """
    center = geometry.circumcenter_lapack(points)
    actual = geometry.circumradius(points, center)
    assert within_range(actual, expected, 0.02)


# Alpha Shape tests below
# Credit for the coordinates utilized in these tests goes to
# user 'bellokk' on git hub (https://github.com/bellockk/alphashape/tree/master)


def test_given_a_point_return_a_point():
    """
    Given a point, the alphashape function should return the same point
    """
    point_a, _ = geometry.concave_hull([(0.0, 0.0)], 0)
    point_c, _ = geometry.concave_hull([(1.0, 0.0)], 0)
    point_b, _ = geometry.concave_hull([(0.0, 1.0)], 0)
    point_d, _ = geometry.concave_hull([(0.0, 0.0)], 99)
    point_e, _ = geometry.concave_hull([(1.0, 0.0)], 99)
    point_f, _ = geometry.concave_hull([(0.0, 1.0)], 99)
    assert Point([0.0, 0.0]) == point_a
    assert Point([1.0, 0.0]) == point_c
    assert Point([0.0, 1.0]) == point_b
    assert Point([0.0, 0.0]) == point_d
    assert Point([1.0, 0.0]) == point_e
    assert Point([0.0, 1.0]) == point_f


def test_given_a_line_with_dupicate_points_return_a_point():
    """
    Given a line with duplicate points, the alphashape function should
    return a point
    """
    actual, _ = geometry.concave_hull([(0.0, 1.0), (0.0, 1.0)], 0)
    assert Point([0.0, 1.0]) == actual


def test_given_a_line_with_unique_points_return_a_line():
    """
    Given a line with unique points, the alphashape function should return
    the same line
    """
    actual_horz, _ = geometry.concave_hull([(0.0, 0.0), (0.0, 1.0)], 0)
    actual_diag, _ = geometry.concave_hull([(1.0, 0.0), (0.0, 1.0)], 0)
    assert LineString([(0.0, 0.0), (0.0, 1.0)]) == actual_horz
    assert LineString([(1.0, 0.0), (0.0, 1.0)]) == actual_diag


def test_given_a_triangle_with_duplicate_points_returns_a_point():
    """
    Given a triangle with two unique points, the alphashape function should
    return a point
    """
    dupe_tri, _ = geometry.concave_hull([(0.0, 1.0), (0.0, 1.0), (0.0, 1.0)], 0)
    assert Point((0.0, 1.0)) == dupe_tri


def test_given_a_triangle_with_two_duplicate_points_returns_a_line():
    """
    Given a line with two unique points, the alphashape function should
    return a line with the unique points
    """
    tri_two_dupe, _ = geometry.concave_hull([(1.0, 0.0), (0.0, 1.0), (0.0, 1.0)], 0)
    assert LineString([(1.0, 0.0), (0.0, 1.0)]) == tri_two_dupe


def test_star_poly():
    """
    Given a 3-dimensional data set, return an expected set of edges.
    """
    coords_2d = sorted(
        [
            (0.0, 0.0),
            (0.0, 0.0),
            (0.0, 1.0),
            (1.0, 0.0),
            (1.0, 1.0),
            (1.0, 0.0),
            (0.0, 1.0),
            (1.0, 1.0),
            (0.25, 0.5),
            (0.5, 0.25),
            (0.5, 0.5),
            (0.75, 0.5),
            (0.5, 0.75),
            (0.5, 0.5),
        ]
    )
    points_2d = [Point((x, y)) for (x, y) in coords_2d]
    # expected_vertices = [
    #     [0.0, 0.0],
    #     [0.0, 0.0],
    #     [0.0, 1.0],
    #     [0.0, 1.0],
    #     [1.0, 0.0],
    #     [1.0, 0.0],
    #     [1.0, 1.0],
    #     [1.0, 1.0],
    #     [0.25, 0.5],
    #     [0.5, 0.25],
    #     [0.5, 0.5],
    #     [0.5, 0.5],
    #     [0.5, 0.75],
    #     [0.75, 0.5],
    # ]
    expected = star_poly
    results, _, _ = geometry.concave_hull(points_2d, 2.1)
    assert results.contains(expected)
    assert expected.contains(results)
    # breakpoint()


@pytest.mark.parametrize(
    "point,points,num, expected",
    [((2, 2), example_coords, 1, (1, 2)), ((9, 7), start_poly_coords, 1, (1, 1))],
)
def test_closest_point(point, points, num, expected):
    actual = geometry.closest_points(point, points, num)
    assert actual == expected



from __future__ import annotations

import logging
import os
from pathlib import Path

from canopyhydro.configuration import input_dir
from canopyhydro.CylinderCollection import CylinderCollection

log = logging.getLogger("model")


NAME = "Forester"


class Forester:
    def __init__(self, directory=input_dir) -> None:
        self.file_names = []
        self.directory = directory
        self.cylinder_collections = []
        self.file_names = self.get_file_names()

    def get_file_names(self):
        directory = self.directory
        log.debug(f"Searching {directory} for files")
        directory = Path(directory)
        paths = sorted(directory.iterdir(), key=os.path.getmtime)
        self.file_names = paths
        file_names = [f.name for f in paths if f.suffix == ".csv"]
        log.debug(f"The following files found in {directory}: {file_names}")
        return paths

    def qsm_to_collection(self, file_name: str = "All", directory: Path = ""):
        """Creates a Cylinder collection from the given file
            and adds the collection to the Forester object

        Args:
            file_name (str):
                Name of the csv file containing QSM data for
                the desired cylinder collection
            directory (str, optional):
                File directory in which the specified file is located.
                Defaults to input_dir.
        """
        if directory == "":
            directory = self.directory
        if self.file_names == "":
            self.get_file_names()
        collections = []
        for file_obj in self.file_names:
            if ".csv" not in file_name and file_name != "All":
                file_name = file_name + ".csv"
            if file_name == "All" or file_obj.name == file_name:
                c = CylinderCollection()
                c.from_csv(file_obj)
                collections.append(c)
        if len(collections) == 0:
            msg = f"File {file_name} not found in input directory {directory}"
            log.error(msg)
            raise FileNotFoundError(msg)
        self.cylinder_collections = collections



from __future__ import annotations

import logging
from typing import ClassVar

from rich.console import Console
from rich.highlighter import Highlighter, ReprHighlighter
from rich.theme import Theme


class ConsoleHandler(logging.StreamHandler):
    HIGHLIGHTER_CLASS: ClassVar[type[Highlighter]] = ReprHighlighter

    def __init__(
        self,
        stream=None,
        highlighter: Highlighter | None = None,
        styles: dict[str, str] = None,
        level: int | str = logging.NOTSET,
    ):
        """
        Allows for logging to the console with rich formatting.
        Args:
            stream (Optional[IO[str]]): A file-like object to write to. Defaults to sys.stderr.
            highlighter (Highlighter, optional): A highlighter to use for syntax
                highlighting. Using  NullHighlighter eliminates highlighting.
            styles (Dict[str, str], optional): A dictionary of styles to use. Defaults to None.
            level (Union[int, str], optional): The minimum log level to write to the console.
                Defaults to logging.NOTSET.

        Note: 'markup = True' allows for customization of error message format based on bracketed tags
                This may cause strange behavior with brackets
        """
        super().__init__(stream=stream)

        highlighter = highlighter or self.HIGHLIGHTER_CLASS()
        theme = Theme(styles, inherit=False)

        self.level = level
        self.console = Console(
            highlighter=highlighter,
            theme=theme,
            file=self.stream,
            markup=True,
        )

    def emit(self, record: logging.LogRecord):
        try:
            message = self.format(record)
            self.console.print(message, soft_wrap=True)
        except Exception as e:
            print(f"Error in console handler: {e}")
